# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Literal

import jax
from jax import numpy as jnp
from jax.sharding import Mesh, NamedSharding
from jax.sharding import PartitionSpec as P

import tpu_inference.envs as envs
from tpu_inference.kernels.gather import gather_reduce as gather_reduce_sc
from tpu_inference.kernels.megablox.gmm_v2 import gmm_v2
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.utils import get_mesh_shape_product


def apply_scoring_fn(scoring_fn: str, x: jax.Array) -> jax.Array:
    match scoring_fn:
        case "softmax":
            return jax.nn.softmax(x, axis=-1)
        case "sigmoid":
            return jax.nn.sigmoid(x)
        case _:
            raise NotImplementedError(
                f"FusedMoE does not support {scoring_fn} scoring function")


def gmm_wrapper(lhs,
                rhs,
                rhs_scale,
                rhs_bias,
                group_sizes,
                group_offset,
                fuse_act=None,
                preferred_element_type=None):
    gmm_res = gmm_v2(
        lhs=lhs,
        rhs=rhs,
        rhs_scale=rhs_scale,
        rhs_bias=rhs_bias,
        group_sizes=group_sizes,
        group_offset=group_offset[0],
        zero_initialize=False,
        fuse_act=fuse_act,
        preferred_element_type=preferred_element_type,
    )
    return gmm_res


def valid_rows_mask(batch_size: int, group_sizes: jax.Array,
                    group_start: jax.Array, group_end: jax.Array) -> jax.Array:
    """Mask indicating rows processed by current shard."""

    group_sizes_sum = jnp.cumulative_sum(group_sizes, include_initial=True)

    token_start = group_sizes_sum[group_start]
    token_end = group_sizes_sum[group_end]

    index = jnp.arange(batch_size)
    return jnp.where(jnp.logical_and(token_start <= index, index < token_end),
                     True, False)


def moe_gmm_local(
    x: jax.Array,
    w1: jax.Array,
    w1_scale: jax.Array | None,
    w1_bias: jax.Array | None,
    w2: jax.Array,
    w2_scale: jax.Array | None,
    w2_bias: jax.Array | None,
    group_sizes: jax.Array,
    group_offset: jax.Array,
    topk_argsort_revert_indices: jax.Array,
    topk_weights: jax.Array,
    *,
    activation: str,
    topk: int,
    parallelism: Literal["tp", "ep"],
    sc_kernel_threshold: int,
    sc_kernel_col_chunk_size: int,
) -> jax.Array:
    """Main MoE logic on a local shard can run in TP or EP mode.

    Set parallelism for "tp" or "ep"
    """

    assert parallelism in ["tp", "ep"]

    # GMM1 computes x @ (W_up | W_gate) together and activation, output is [tokens,padded_intermediate_size]
    gmm1_res = gmm_wrapper(
        x,
        w1,
        w1_scale,
        w1_bias,
        group_sizes,
        group_offset,
        fuse_act=activation,
        preferred_element_type=x.dtype,
    )

    # When the parallelism is TP since w2_bias is not sharded, we should only apply bias
    # once, not applying to every shard. So we set w2_bias to 0 to all shards other than
    # shard 0. For EP, it is not needed since bias is sharded on leading expert axis.
    if parallelism == "tp" and w2_bias is not None:
        shard_id = jax.lax.axis_index(ShardingAxisName.MLP_TENSOR).sum()
        w2_bias = jnp.where(shard_id == 0, w2_bias, 0)
    gmm1_res = gmm1_res[:, :w2.shape[1]]  # trim to hidden size if padded

    local_group_size = w1.shape[0]
    if local_group_size < group_sizes.size:
        mask = valid_rows_mask(
            gmm1_res.shape[0],
            group_sizes,
            group_offset,
            group_offset + local_group_size,
        )[topk_argsort_revert_indices]

    if gather_reduce_sc.is_supported_by_sc_gather_reduce(
            gmm1_res.shape[0], sc_kernel_threshold):
        gmm2_res = gmm_wrapper(gmm1_res,
                               w2,
                               w2_scale,
                               w2_bias,
                               group_sizes,
                               group_offset,
                               preferred_element_type=jnp.float32.dtype)

        if local_group_size < group_sizes.size:
            mask = mask.reshape(-1, topk)
            topk_weights = jnp.where(mask, topk_weights, 0)

        inds = topk_argsort_revert_indices
        topk_weights = topk_weights.flatten().reshape(-1, 128)

        token_hidden = gather_reduce_sc.sc_gather_reduce(
            op=gmm2_res,
            idx=inds,
            reduce_group_size=topk,
            topk_weights=topk_weights,
            col_chunk_size=sc_kernel_col_chunk_size,
        )
    else:
        gmm2_res = gmm_wrapper(gmm1_res,
                               w2,
                               w2_scale,
                               w2_bias,
                               group_sizes,
                               group_offset,
                               preferred_element_type=x.dtype)

        # First run local reduction on topk experts owned by the rank for all tokens
        token_topk_hidden = gmm2_res[topk_argsort_revert_indices].reshape(
            (-1, topk, gmm2_res.shape[-1]))
        token_topk_hidden = token_topk_hidden * jnp.expand_dims(topk_weights,
                                                                axis=-1)

        if local_group_size < group_sizes.size:
            mask = mask.reshape(-1, topk, 1)
            token_topk_hidden = jnp.where(mask, token_topk_hidden, 0.0)

        token_hidden = token_topk_hidden.sum(axis=-2)

    reduction_axis = (ShardingAxisName.MLP_TENSOR
                      if parallelism == "tp" else ShardingAxisName.EXPERT)
    # Then global reduction on all ranks for all tokens and all experts
    return jax.lax.psum(token_hidden, axis_name=reduction_axis).astype(x.dtype)


def tensor_parallel_gmm(
    x: jax.Array,
    w1: jax.Array,
    w1_scale: jax.Array | None,
    w1_bias: jax.Array | None,
    w2: jax.Array,
    w2_scale: jax.Array | None,
    w2_bias: jax.Array | None,
    group_sizes: jax.Array,
    topk_argsort_revert_indices: jax.Array,
    topk_weights: jax.Array,
    *,
    activation: str,
    topk: int,
    mesh: Mesh,
    sc_kernel_threshold: int,
    sc_kernel_col_chunk_size: int,
) -> jax.Array:
    data_p_spec = P(ShardingAxisName.MLP_DATA)
    group_offset = jnp.array([0])

    w1_spec = P(None, None, ShardingAxisName.MLP_TENSOR)
    w2_spec = P(None, ShardingAxisName.MLP_TENSOR, None)

    w1_scale_spec = (None if w1_scale is None else P(
        None, None, None, ShardingAxisName.MLP_TENSOR))
    w1_bias_spec = (None if w1_bias is None else P(
        None, None, ShardingAxisName.MLP_TENSOR))

    num_blocks = 1 if w2_scale is None else w2_scale.shape[1]
    w2_scale_spec = (None if num_blocks == 1 else P(
        None, ShardingAxisName.MLP_TENSOR, None, None))
    w2_bias_spec = None if w2_bias is None else P(None, None, None)

    return jax.shard_map(
        functools.partial(
            moe_gmm_local,
            activation=activation,
            topk=topk,
            parallelism="tp",
            sc_kernel_threshold=sc_kernel_threshold,
            sc_kernel_col_chunk_size=sc_kernel_col_chunk_size,
        ),
        mesh=mesh,
        in_specs=(
            data_p_spec,
            w1_spec,
            w1_scale_spec,
            w1_bias_spec,
            w2_spec,
            w2_scale_spec,
            w2_bias_spec,
            data_p_spec,
            P(),
            data_p_spec,
            data_p_spec,
        ),
        out_specs=(data_p_spec),
        check_vma=False,
    )(
        x,
        w1,
        w1_scale,
        w1_bias,
        w2,
        w2_scale,
        w2_bias,
        group_sizes,
        group_offset,
        topk_argsort_revert_indices,
        topk_weights,
    )


def expert_parallel_gmm(
    x: jax.Array,
    w1: jax.Array,
    w1_scale: jax.Array | None,
    w1_bias: jax.Array | None,
    w2: jax.Array,
    w2_scale: jax.Array | None,
    w2_bias: jax.Array | None,
    group_sizes: jax.Array,
    topk_argsort_revert_indices: jax.Array,
    topk_weights: jax.Array,
    *,
    activation: str,
    topk: int,
    mesh: Mesh,
    sc_kernel_threshold: int,
    sc_kernel_col_chunk_size: int,
) -> jax.Array:
    ep_size = get_mesh_shape_product(mesh, ShardingAxisName.EXPERT)
    ep_p_spec = P(ShardingAxisName.EXPERT)
    data_p_spec = P(ShardingAxisName.MLP_DATA)
    num_experts = w1.shape[0]
    num_experts_per_shard = num_experts // ep_size
    group_offset = jnp.arange(0, num_experts, num_experts_per_shard)

    w1_scale_spec = None if w1_scale is None else ep_p_spec
    w1_bias_spec = None if w1_bias is None else ep_p_spec
    w2_scale_spec = None if w2_scale is None else ep_p_spec
    w2_bias_spec = None if w2_bias is None else ep_p_spec

    return jax.shard_map(
        functools.partial(
            moe_gmm_local,
            activation=activation,
            topk=topk,
            parallelism="ep",
            sc_kernel_threshold=sc_kernel_threshold,
            sc_kernel_col_chunk_size=sc_kernel_col_chunk_size,
        ),
        mesh=mesh,
        in_specs=(
            data_p_spec,
            ep_p_spec,
            w1_scale_spec,
            w1_bias_spec,
            ep_p_spec,
            w2_scale_spec,
            w2_bias_spec,
            data_p_spec,
            ep_p_spec,
            data_p_spec,
            data_p_spec,
        ),
        out_specs=(data_p_spec),
        check_vma=False,
    )(
        x,
        w1,
        w1_scale,
        w1_bias,
        w2,
        w2_scale,
        w2_bias,
        group_sizes,
        group_offset,
        topk_argsort_revert_indices,
        topk_weights,
    )


@jax.jit(static_argnames=(
    "topk",
    "renormalize",
    "mesh",
    "use_ep",
    "activation",
    "scoring_fn",
    "sc_kernel_threshold",
    "sc_kernel_col_chunk_size",
))
def fused_moe_func(
    hidden_states: jax.Array,
    w1: jax.Array,
    w2: jax.Array,
    w1_scale: jax.Array | None,
    w2_scale: jax.Array | None,
    w1_bias: jax.Array | None,
    w2_bias: jax.Array | None,
    gating_output: jax.Array,
    topk: int,
    renormalize: bool,
    mesh: Mesh,
    use_ep: bool,
    activation: str,
    scoring_fn: str,
    sc_kernel_threshold: int,
    sc_kernel_col_chunk_size: int,
) -> jax.Array:
    """Route tokens in hidden_states into each experts based on routing.

    Args:
        hidden_states: [num_tokens, hidden_size]
        w1: first moe weights [num_experts, hidden_size, intermediate_size * 2]
        w2: second moe weights [num_experts, intermediate_size, hidden_size]
        w1_scale: w1 scale [num_experts, num_blocks, 1, intermediate_size * 2]
        w2_scale: w2 scale [num_experts, num_blocks, 1, hidden_size]
        w1_bias: optional bias of w1 [num_experts, 1, intermediate_size * 2]
        w2_bias: optional bias of w2 [num_experts, 1, hidden_size]
        gating_output: routing information of tokens [num_tokens, num_experts]
        topk: number of experts to choose per token.
        renormalize: normalize gating_output.
        mesh: mesh to perform moe.
        use_ep: use expert parallelism.
        activation: activation function to perform on the output of w1.
        scoring_fn: scoring function to apply on gating_output.

    Returns:
        Output of moe operation [num_tokens, hidden_size]
    """
    num_tokens, hidden_size = hidden_states.shape
    global_num_experts, padded_hidden_size, _ = w1.shape
    dtype = hidden_states.dtype

    assert (num_tokens * topk) % 16 == 0, (
        "The kernel requires num_tokens * topk to be a multiple of "
        f"16 but got {num_tokens}*{topk}={num_tokens*topk}")

    assert gating_output.shape == (num_tokens, global_num_experts)

    topk_weights = apply_scoring_fn(scoring_fn, gating_output)
    if envs.FORCE_MOE_RANDOM_ROUTING:
        # Forcing random routing is useful to get rid of the effect
        # of routing imbalance during performance debugging.
        rng_key = jax.random.PRNGKey(0)
        topk_indices = jax.vmap(lambda key: jax.random.choice(
            key, global_num_experts, shape=(topk, ), replace=False))(
                jax.random.split(rng_key, num_tokens))
        topk_weights = topk_weights[jnp.arange(num_tokens)[:, None],
                                    topk_indices]
    else:
        topk_weights, topk_indices = jax.lax.top_k(topk_weights, k=topk)
    if renormalize:
        topk_weights = topk_weights / topk_weights.sum(axis=-1, keepdims=True)
    topk_weights = topk_weights.astype(dtype)
    # All-gather topk weights for attention dp
    topk_weights = jax.lax.with_sharding_constraint(
        topk_weights, NamedSharding(mesh, P(ShardingAxisName.MLP_DATA, None)))

    def _process_tokens_locally(hidden_states_local, topk_indices_local):
        num_tokens_local = hidden_states_local.shape[0]
        topk_indices_flat = topk_indices_local.flatten()
        topk_argsort_indices = jnp.argsort(topk_indices_flat)
        token_indices = jnp.arange(num_tokens_local,
                                   dtype=jnp.int32).repeat(topk)
        token_indices_sorted = token_indices[topk_argsort_indices]
        x = hidden_states_local[token_indices_sorted]
        # Below one_hot is equivalent to jnp.bincount(topk_indices_flat,
        # length=global_num_experts) but is more performant.
        group_sizes_local = jax.nn.one_hot(topk_indices_flat,
                                           global_num_experts,
                                           dtype=jnp.int32).sum(axis=0)
        topk_argsort_revert_indices = jnp.argsort(topk_argsort_indices)

        return x, group_sizes_local, topk_argsort_revert_indices

    x, group_sizes, topk_argsort_revert_indices = jax.shard_map(
        _process_tokens_locally,
        mesh=mesh,
        in_specs=(
            P(ShardingAxisName.MLP_DATA, None),
            P(ShardingAxisName.MLP_DATA, None),
        ),
        out_specs=(
            P(ShardingAxisName.MLP_DATA, None),
            P(ShardingAxisName.MLP_DATA),
            P(ShardingAxisName.MLP_DATA),
        ),
    )(hidden_states, topk_indices)

    x = jnp.pad(x, ((0, 0), (0, padded_hidden_size - hidden_size)))

    if use_ep:
        x = expert_parallel_gmm(
            x,
            w1,
            w1_scale,
            w1_bias,
            w2,
            w2_scale,
            w2_bias,
            group_sizes,
            topk_argsort_revert_indices,
            topk_weights,
            activation=activation,
            topk=topk,
            mesh=mesh,
            sc_kernel_threshold=sc_kernel_threshold,
            sc_kernel_col_chunk_size=sc_kernel_col_chunk_size,
        )
    else:
        x = tensor_parallel_gmm(
            x,
            w1,
            w1_scale,
            w1_bias,
            w2,
            w2_scale,
            w2_bias,
            group_sizes,
            topk_argsort_revert_indices,
            topk_weights,
            activation=activation,
            topk=topk,
            mesh=mesh,
            sc_kernel_threshold=sc_kernel_threshold,
            sc_kernel_col_chunk_size=sc_kernel_col_chunk_size,
        )

    return x[:num_tokens, :hidden_size]



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from abc import ABC, abstractmethod
from typing import Optional

import jax

from tpu_inference.layers.common.quantization.configs import \
    QuantLinearConfig as CommonQuantLinearConfig
from tpu_inference.layers.jax import JaxModule
from tpu_inference.layers.jax.quantization import QuantizeMethodBase


class QuantizationConfig(ABC):

    def __init__(self, hf_quant_config: dict):
        pass

    @abstractmethod
    def get_quant_method(self, layer: JaxModule,
                         prefix: str) -> Optional[QuantizeMethodBase]:
        raise NotImplementedError

    @classmethod
    def get_from_keys(cls, config: dict, keys: list, *args):
        """Get value from config using the first matching key.'
        
        Return default value if no key is found and default is provided.
        Raise KeyError if no key is found and no default is provided.
        """
        assert len(args) <= 1, "Only one default value is allowed."
        for key in keys:
            if key in config:
                return config[key]
        if args:
            return args[0]
        raise KeyError(f"None of the keys {keys} found in config.")

    @classmethod
    def is_layer_skipped(
        cls,
        prefix: str,
        *,
        ignored_layers: list[str],
        fused_mapping: dict = dict()) -> bool:
        """Check if a layer should be skipped from quantization.

        Follows: https://github.com/vllm-project/vllm/blob/main/vllm/model_executor/layers/quantization/utils/quant_utils.py#L418
        """

        def prefix_full_match(prefix: str, ignored_layers: list[str]) -> bool:
            return prefix in ignored_layers

        match_func = prefix_full_match

        proj_name = prefix.split(".")[-1]

        if fused_mapping and proj_name in fused_mapping:
            shard_prefixes = [
                prefix.replace(proj_name, shard_proj_name)
                for shard_proj_name in fused_mapping[proj_name]
            ]

            is_skipped = None
            for shard_prefix in shard_prefixes:
                is_shard_skipped = match_func(shard_prefix, ignored_layers)

                if is_skipped is None:
                    is_skipped = is_shard_skipped
                elif is_shard_skipped != is_skipped:
                    raise ValueError(
                        f"Detected some but not all shards of {prefix} "
                        "are quantized. All shards of fused layers "
                        "must have the same precision.")
        elif "experts" in prefix:
            expert_ignore_layers = [
                layer_name for layer_name in ignored_layers
                if "experts" in layer_name
            ]
            is_skipped = any(prefix in layer_name
                             for layer_name in expert_ignore_layers)
        else:
            is_skipped = match_func(prefix, ignored_layers)

        return is_skipped


def _to_partition_spec(sharding) -> jax.sharding.PartitionSpec:
    """Convert a sharding value to a PartitionSpec.

    Handles NamedSharding (extracts .spec), raw tuples/lists from
    nnx.with_partitioning, and passthrough for existing PartitionSpec.
    """
    if isinstance(sharding, jax.sharding.NamedSharding):
        return sharding.spec
    if isinstance(sharding, jax.sharding.PartitionSpec):
        return sharding
    if isinstance(sharding, (tuple, list)):
        return jax.sharding.PartitionSpec(*sharding)
    return jax.sharding.PartitionSpec()


class QuantLinearConfig(CommonQuantLinearConfig):
    """Quantization config for jax linear layers."""

    def __init__(self, layer, *, enable_sp: bool):
        # Avoid circular import.
        from tpu_inference.layers.jax.linear import JaxEinsum
        assert isinstance(layer, JaxEinsum)
        # Update config attributes by parsing einsum string and weight sharding.
        # Parse the einsum string to classify axes:
        #   - contracting: in both operands but NOT in output (summed over)
        #   - batch: in both operands AND in output (paired/indexed)
        #   - free: in only one operand and in output
        einsum_str = layer.einsum_str
        weight = layer.weight

        lhs, output_axis = einsum_str.replace(" ", "").split("->")
        x_axis, w_axis = lhs.split(",")

        shared_axes = set(x_axis) & set(w_axis)
        batch_axes = shared_axes & set(output_axis)
        contracting_axes = shared_axes - batch_axes

        self.in_features = tuple(weight.shape[i] for i, c in enumerate(w_axis)
                                 if c in contracting_axes)

        # Extract and fuse sharding per axis category.
        sharding = _to_partition_spec(getattr(weight, "sharding", ()))
        sharding = sharding + (None, ) * (len(weight.shape) - len(sharding))

        in_sharding = set(s for i, s in enumerate(sharding)
                          if w_axis[i] in contracting_axes and s is not None)
        out_sharding = set(
            s for i, s in enumerate(sharding)
            if w_axis[i] not in (contracting_axes
                                 | batch_axes) and s is not None)
        batch_sharding_set = set(s for i, s in enumerate(sharding)
                                 if w_axis[i] in batch_axes and s is not None)

        assert len(in_sharding) <= 1 and len(out_sharding) <= 1, \
            f"Cannot fuse sharding {getattr(weight, 'sharding', ())=} into 2D weight sharding for {einsum_str}"

        self.out_features = tuple(
            weight.shape[i] for i, c in enumerate(w_axis)
            if c not in contracting_axes and c not in batch_axes)
        self.batch_features = tuple(weight.shape[i]
                                    for i, c in enumerate(w_axis)
                                    if c in batch_axes)

        self.out_features_sharding = (next(iter(out_sharding), None), )
        self.in_features_sharding = (next(iter(in_sharding), None), )
        self.batch_sharding = tuple(batch_sharding_set)

        output_sizes = [math.prod(self.out_features)]
        super().__init__(enable_sp=enable_sp, output_sizes=output_sizes)

        # Update weight_sharding and bias_sharding for 2D matmul compatibility
        if self.batch_features:
            self.weight_sharding = _to_partition_spec(
                weight.get_metadata().get("sharding", ()))
        else:
            self.weight_sharding = jax.sharding.PartitionSpec(
                *(self.out_features_sharding + self.in_features_sharding))

        self.bias_sharding = jax.sharding.PartitionSpec(
            *self.out_features_sharding)



# SPDX-License-Identifier: Apache-2.0
"""Utility functions for quantized matmul kernel."""
from typing import Any, Callable

import jax
import jax.numpy as jnp
from jax._src import dtypes

from tpu_inference.kernels.quantized_matmul.tuned_block_sizes import TunedValue


def unfold_args(
    conditions: tuple[jax.Array | bool, ...],
    fn_conditions: tuple[bool, ...],
    fn: Callable[..., Any],
):
    """Minimize run-time branching of fn by converting jnp.bool to python bool."""
    if conditions:
        arg = conditions[0]
        if isinstance(arg, bool):
            unfold_args(conditions[1:], fn_conditions + (arg, ), fn)
        else:
            assert arg.dtype == jnp.bool and arg.size == 1
            jax.lax.cond(
                arg,
                lambda: unfold_args(conditions[1:], fn_conditions +
                                    (True, ), fn),
                lambda: unfold_args(conditions[1:], fn_conditions +
                                    (False, ), fn),
            )
    else:
        fn(*fn_conditions)


def quantize_tensor(x: jax.Array,
                    dtype: jnp.dtype,
                    dim: int = -1,
                    block_size: int | None = None):
    if block_size is not None:
        # Flatten all leading dims into a single batch dim for block
        # quantization, then restore the original shape.
        orig_shape = x.shape
        k_dim = orig_shape[-1]
        x_flat = x.reshape(-1, k_dim)
        n_dim = x_flat.shape[0]
        x_reshaped = x_flat.reshape(n_dim, -1, block_size)
        x_q, scale = quantize_block(x_reshaped, axis=-1, target_dtype=dtype)

        x_q = x_q.reshape(orig_shape)

        return x_q, scale.transpose(1, 2, 0).astype(jnp.float32)
    data_q, scale = quantize_block(x, axis=dim, target_dtype=dtype)
    return data_q, scale.astype(jnp.float32)


def next_multiple(x, multiple):
    return ((x + multiple - 1) // multiple) * multiple


def get_kernel_name(tuned_value: TunedValue):
    batch_block_size = tuned_value.batch_block_size
    out_block_size = tuned_value.out_block_size
    in_block_size = tuned_value.in_block_size
    return (
        f"quantized_matmul_kernel_{batch_block_size}_{out_block_size}_{in_block_size}"
    )


def xla_quantized_matmul(
    x: jax.Array,
    w_q: jax.Array,
    w_scale: jax.Array,
    quantize_activation=True,
) -> jax.Array:
    """
    Reference (pure JAX) implementation of the quantized matmul kernel below.

    Args:
        x:  Activation.
        w_q: Weight quantized array. [n_output_features, n_input_features]
        w_s: Weight quantization scale. [n_output_features]
        mesh: Mesh to shard on.
        weight_sharding: PartitionSpec for the weight tensor.

    Returns:
        Output of the quantized matmul.
    """
    if quantize_activation:
        acc_dtype = jnp.float32
        if quantize_activation and jnp.issubdtype(w_q.dtype, jnp.integer):
            acc_dtype = jnp.int32

        x_q, x_scale = quantize_tensor(x, w_q.dtype)
        out = jax.lax.dot_general(
            x_q,
            w_q,
            dimension_numbers=(((1, ), (1, )), ((), ())),
            preferred_element_type=acc_dtype,
        ).astype(jnp.float32)
        out *= x_scale
    else:
        out = jax.lax.dot_general(
            x,
            w_q,
            dimension_numbers=(((1, ), (1, )), ((), ())),
            preferred_element_type=jnp.float32,
        )
    out *= jnp.expand_dims(w_scale, 0)
    return out.astype(x.dtype)


def xla_quantized_batched_matmul(
    x: jax.Array,
    w_q: jax.Array,
    w_scale: jax.Array,
    dimension_numbers: tuple,
    quantize_activation: bool = True,
) -> jax.Array:
    """Quantized matmul with batch dimensions via dot_general.

    Generalizes xla_quantized_matmul to support batch dimensions (axes
    shared between both operands that appear in the output). Uses
    jax.lax.dot_general with configurable dimension_numbers.

    Args:
        x: Activation tensor (e.g. [T, N, H]).
        w_q: Quantized weight (e.g. [A, N, H]).
        w_scale: Per-output-channel weight scale (e.g. [A]).
        dimension_numbers: ``((contracting_x, contracting_w),
            (batch_x, batch_w))`` for dot_general.
        quantize_activation: Whether to dynamically quantize activations.

    Returns:
        Output of the batched quantized matmul. Shape is determined by
        dot_general: batch dims first, then lhs free dims, then rhs free dims.
    """
    contract_dims, batch_dims = dimension_numbers

    if quantize_activation:
        acc_dtype = jnp.float32
        if jnp.issubdtype(w_q.dtype, jnp.integer):
            acc_dtype = jnp.int32

        x_q, x_scale = quantize_tensor(x, w_q.dtype)
        out = jax.lax.dot_general(
            x_q,
            w_q,
            dimension_numbers=(contract_dims, batch_dims),
            preferred_element_type=acc_dtype,
        ).astype(jnp.float32)
        # Permute x_scale to match dot_general output ordering.
        # dot_general output: batch dims, lhs free dims, rhs free dims.
        # x_scale has same shape as x but with contracting dim(s) as 1.
        contract_set = set(contract_dims[0])
        batch_set = set(batch_dims[0])
        lhs_free = [
            i for i in range(x.ndim)
            if i not in contract_set and i not in batch_set
        ]
        perm = list(batch_dims[0]) + lhs_free + list(contract_dims[0])
        if perm != list(range(x.ndim)):
            x_scale = jnp.transpose(x_scale, perm)
        out *= x_scale
    else:
        out = jax.lax.dot_general(
            x,
            w_q,
            dimension_numbers=(contract_dims, batch_dims),
            preferred_element_type=jnp.float32,
        )

    # Broadcast w_scale to match the output shape.
    # dot_general output order: batch dims, lhs free dims, rhs free dims.
    # w_scale is per-output-channel (rhs free dims), so expand leading dims.
    n_leading = out.ndim - w_scale.ndim
    for _ in range(n_leading):
        w_scale = jnp.expand_dims(w_scale, 0)
    out *= w_scale
    return out.astype(x.dtype)


def quantize_array(
    x: jax.Array,  # [bs_block_size, in_block_size]
    x_abs_max: jax.Array,  # [1, bs_block_size]
    quant_dtype: jnp.dtype,
):
    is_float = jnp.issubdtype(quant_dtype, jnp.floating)
    dtype_info = jnp.finfo(quant_dtype) if is_float else jnp.iinfo(quant_dtype)
    dtype_max = float(dtype_info.max)

    # TODO(kyuyeunk): Investigate performance gain from non xlu transpose.
    scale = jnp.transpose(x_abs_max / dtype_max)
    scale_inv = jnp.nan_to_num(1 / scale, dtype_max)
    return (x * scale_inv).astype(quant_dtype), scale.astype(jnp.float32)


def get_vmem_limit(
    n_batch: int,
    n_out: int,
    n_in: int,
    batch_block_size: int,
    out_block_size: int,
    in_block_size: int,
    x_dtype: jnp.dtype,
    x_q_dtype: jnp.dtype,
    w_q_dtype: jnp.dtype,
    scale_dtype: jnp.dtype,
    out_dtype: jnp.dtype,
    acc_dtype: jnp.dtype,
    save_acc: bool,
    save_x_q: bool,
    upper_limit_bytes: int,
):
    """Calculate VMEM limit for the kernel."""

    # Calculate in/out VMEM size.
    x_size = (batch_block_size * in_block_size * dtypes.itemsize_bits(x_dtype))
    x_abs_max_size = batch_block_size * dtypes.itemsize_bits(scale_dtype)
    w_q_size = (out_block_size * in_block_size *
                dtypes.itemsize_bits(w_q_dtype))
    w_scale_size = out_block_size * dtypes.itemsize_bits(scale_dtype)
    out_size = (batch_block_size * out_block_size *
                dtypes.itemsize_bits(out_dtype))

    vmem_in_out = x_size + x_abs_max_size + w_q_size + w_scale_size + out_size
    vmem_in_out *= 2  # Account for compute and vreg spills.

    # Account for double buffering.
    # Double buffering is used only if there are multiple blocks per in/out.
    vmem_in_out += x_size if (n_batch > 1 or n_in > 1) else 0
    vmem_in_out += x_abs_max_size if (n_batch > 1) else 0
    vmem_in_out += w_q_size if (n_out > 1 or n_in > 1) else 0
    vmem_in_out += w_scale_size if (n_out > 1) else 0
    vmem_in_out += out_size if (n_batch > 1 or n_out > 1) else 0

    # Calculate scratch VMEM size.
    acc_size = (batch_block_size * out_block_size *
                dtypes.itemsize_bits(acc_dtype))
    x_q_size = (batch_block_size * in_block_size *
                dtypes.itemsize_bits(x_q_dtype))
    x_scale_size = batch_block_size * dtypes.itemsize_bits(scale_dtype)

    vmem_scratch = acc_size if save_acc else 0
    vmem_scratch += x_q_size + x_scale_size if save_x_q else 0
    vmem_scratch *= 2  # Account for compute and vreg spills.

    # Add in/out and scratch VMEM size.
    vmem_used = vmem_in_out + vmem_scratch
    vmem_used_bytes = vmem_used // 8  # Convert bits to bytes.
    # Specify upper limit. Defaults to 96MB.
    vmem_limit_bytes = min(vmem_used_bytes, upper_limit_bytes)

    return vmem_limit_bytes


def validate_inputs(
    x: jax.Array,
    w_q: jax.Array,
    w_scale: jax.Array,
    x_abs_max: jax.Array,
    batch_block_size: int,
    out_block_size: int,
    in_block_size: int,
):
    """Verify inputs invoking the kernel."""

    # Verify input shapes.
    if x.shape[1] != w_q.shape[1]:
        raise ValueError(f'{x.shape[1]=} must be equal to {w_q.shape[1]=}')
    if w_q.shape[0] != w_scale.shape[1] and (w_scale.ndim == 3 and w_q.shape[0]
                                             != w_scale.shape[2]):
        raise ValueError(
            f"{w_q.shape[0]=} must be equal to {w_scale.shape[1]=}")
    if x_abs_max is not None and x_abs_max.shape != (1, x.shape[0]):
        raise ValueError(
            f"{x_abs_max.shape=} must be equal to (1, {x.shape[0]=})")
    if x.shape[0] % batch_block_size != 0:
        raise ValueError(
            f"{x.shape[0]=} must be a multiple of {batch_block_size=}")
    if w_q.shape[0] % out_block_size != 0:
        raise ValueError(
            f"{w_q.shape[0]=} must be a multiple of {out_block_size=}")
    if x.shape[1] % in_block_size != 0:
        raise ValueError(
            f"{x.shape[1]=} must be a multiple of {in_block_size=}")


def get_max_min(target_dtype):
    if jnp.issubdtype(target_dtype, jnp.floating):
        return jnp.finfo(target_dtype).max.astype(
            jnp.float32), jnp.finfo(target_dtype).min.astype(jnp.float32)
    else:
        return jnp.iinfo(target_dtype).max, jnp.iinfo(target_dtype).min


def quantize_block(data, axis, target_dtype):
    """Calculates scale and quantizes a block of data."""
    abs_max = jnp.max(
        jnp.abs(data),
        axis=axis,
        keepdims=True,
    )
    dtype_max, dtype_min = get_max_min(target_dtype)
    scale = abs_max / dtype_max
    scale = jnp.where(scale == 0, 1.0, scale)

    if jnp.issubdtype(target_dtype, jnp.floating):
        data_q = (data / scale).clip(dtype_min, dtype_max).astype(target_dtype)
    else:
        data_q = jnp.round(data / scale).astype(target_dtype)
    return data_q, scale



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import random
import string
import time

import pytest
from vllm import LLM, SamplingParams


@pytest.fixture
def sampling_config():
    return SamplingParams(temperature=0,
                          max_tokens=120,
                          ignore_eos=True,
                          repetition_penalty=1,
                          frequency_penalty=0,
                          presence_penalty=0,
                          min_p=0,
                          logprobs=None)


@pytest.fixture
def model_name():
    return "Qwen/Qwen2.5-1.5B-Instruct"


def get_test_prompts():
    """
    Generates a list of prompts with a specific word count,

    Args:
        num_prompts: The number of prompts to generate.
        input_len_words: The total number of words for each prompt.

    Returns:
        A list of strings with number of prompts = num_prompts and
        The total number of words for each prompt = input_len_words.
    """
    num_prompts = 500
    input_len_words = 120
    prompts = []

    # For example w = 's'
    # The generated prompt will be Keep repeating: s s s ...
    num_repetitions = input_len_words
    prefix = "Keep repeating: "

    for _ in range(num_prompts):
        # 1. Pick a random lowercase letter
        w = random.choice(list(string.ascii_lowercase))

        # 2. Create the string of repeated words
        #    This will have (num_repetitions) words
        repeating_part = " ".join([w] * num_repetitions)

        # 3. Combine with the prefix (if any)
        print(f"{prefix}{repeating_part}")
        prompts.append(f"{prefix}{repeating_part}")

    return prompts


def _test_performance_helper(monkeypatch: pytest.MonkeyPatch,
                             sampling_config: SamplingParams, model_name: str,
                             min_speedup: float):
    '''
    Helper function to test async scheduler decoding performance.
    Compares timing between reference LLM and async LLM using Qwen2.5-1.5B.
    '''

    with monkeypatch.context():
        # Use a smaller set of prompts for performance testing
        test_prompts = get_test_prompts()  # num_prompts=100, input_len=120

        # Test reference LLM timing
        ref_llm = LLM(model=model_name,
                      max_model_len=800,
                      max_num_seqs=24,
                      max_num_batched_tokens=512,
                      enable_prefix_caching=False,
                      async_scheduling=0)

        start_time = time.time()
        _ = ref_llm.generate(test_prompts, sampling_config)
        ref_time = time.time() - start_time

        del ref_llm
        # Waiting for TPUs to be released
        time.sleep(10)

        # # Test async LLM timing with max_num_seqs=256
        async_llm = LLM(model=model_name,
                        max_model_len=800,
                        max_num_seqs=24,
                        max_num_batched_tokens=512,
                        enable_prefix_caching=False,
                        async_scheduling=1)

        start_time = time.time()
        _ = async_llm.generate(test_prompts, sampling_config)
        async_time = time.time() - start_time

        del async_llm
        # # Waiting for TPUs to be released
        time.sleep(10)

        speedup = ref_time / async_time
        print(f"Reference LLM time: {ref_time:.2f}s")
        print(f"Async LLM time: {async_time:.2f}s")
        print(f"Speedup: {speedup:.2f}x")

        assert speedup >= min_speedup, f"Expected at least {min_speedup}x speedup for async scheduler, got {speedup:.2f}x"


def test_performance(
    monkeypatch: pytest.MonkeyPatch,
    sampling_config: SamplingParams,
    model_name: str,
):
    '''
    Test that async scheduler decoding provides significant performance improvement.
    Compares timing between reference LLM and async LLM using Qwen2.5-1.5B.
    Expects async_llm to be at least 1.1x faster than ref_llm.
    '''
    min_speed_up = 1.1  # Adjusted the performance thresholds for this specific test to better reflect the realities of the CI environment
    _test_performance_helper(monkeypatch, sampling_config, model_name,
                             min_speed_up)


def _test_correctness_helper(
    monkeypatch: pytest.MonkeyPatch,
    sampling_config: SamplingParams,
    model_name: str,
):
    '''
    Helper function to test async scheduler correctness.
    Compare the outputs of a original LLM and a async LLM
    should be the same when using async scheduler decoding.

    Known Edge Case (KV Cache Swapping):
    Under this case, though the temperature is set to 0,
    the output is still slightly different everytime.
    This is an expected behaviour as the normal scheduler also
    behaves the same and hence, it is difficult to design a test
    for such scenario.
    '''
    with monkeypatch.context():
        test_prompts = get_test_prompts()

        ref_llm = LLM(model=model_name,
                      max_model_len=1024,
                      max_num_seqs=100,
                      async_scheduling=0)
        ref_outputs = ref_llm.generate(test_prompts, sampling_config)

        del ref_llm

        # Waiting for TPUs to be released.
        time.sleep(10)

        async_llm = LLM(model=model_name,
                        max_model_len=1024,
                        max_num_seqs=100,
                        async_scheduling=1)
        async_outputs = async_llm.generate(test_prompts, sampling_config)

        matches = 0
        misses = 0
        for ref_output, async_output in zip(ref_outputs, async_outputs):
            if ref_output.outputs[0].text == async_output.outputs[0].text:
                print(f"ref_output: {ref_output.outputs[0].text}")
                print(f"async_output: {async_output.outputs[0].text}")
                matches += 1
            else:
                misses += 1
                print(
                    f"missed ref_output: {ref_output.outputs[0].text} \n missed ref_output ends"
                )
                print(
                    f"missed async_output: {async_output.outputs[0].text} \n missed async_output ends"
                )

        assert misses <= 5  # Adjusted the performance thresholds for this specific test to better reflect the realities of the CI environment
        del async_llm

        # Waiting for TPUs to be released.
        time.sleep(10)


def test_async_correctness(
    monkeypatch: pytest.MonkeyPatch,
    sampling_config: SamplingParams,
    model_name: str,
):
    '''
    Compare the outputs of a original LLM and a async LLM
    should be the same when using async scheduler.
    '''

    _test_correctness_helper(monkeypatch, sampling_config, model_name)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
from vllm.config import (CacheConfig, ModelConfig, ParallelConfig,
                         SchedulerConfig, SpeculativeConfig, VllmConfig)

from tpu_inference.runner.tpu_runner import TPUModelRunner


class TestTPUJaxRunner:

    def setup_method(self):
        # Mock JAX dependencies
        self.mock_devices = [MagicMock(coords=i) for i in range(1)]
        self.mock_rng_key = MagicMock()
        device_array = np.array(jax.devices()[:1]).reshape(1, 1, 1, -1)
        self.mock_mesh = jax.make_mesh(device_array.shape,
                                       ('data', 'attn_dp', 'expert', 'model'))
        with patch('jax.devices', return_value=self.mock_devices), \
             patch('jax.make_mesh', return_value=self.mock_mesh), \
             patch('jax.random.key', return_value=self.mock_rng_key), \
             patch('tpu_inference.runner.tpu_runner.get_model', return_value=MagicMock()), \
             patch('tpu_inference.runner.tpu_runner.make_optimized_mesh', return_value=self.mock_mesh):

            model_config = ModelConfig(tokenizer_mode="auto",
                                       trust_remote_code=False,
                                       seed=0,
                                       dtype='bfloat16')
            cache_config = CacheConfig(
                block_size=16,
                gpu_memory_utilization=0.9,
                cache_dtype="auto",
            )
            scheduler_config = SchedulerConfig(max_num_seqs=16,
                                               max_model_len=1024,
                                               is_encoder_decoder=False)
            parallel_config = ParallelConfig(
                pipeline_parallel_size=1,
                tensor_parallel_size=1,
            )
            speculative_config = SpeculativeConfig(
                model='ngram',
                num_speculative_tokens=5,
                prompt_lookup_max=4,
            )
            vllm_config = VllmConfig(
                model_config=model_config,
                cache_config=cache_config,
                scheduler_config=scheduler_config,
                parallel_config=parallel_config,
                speculative_config=speculative_config,
                observability_config={},
                additional_config={},
            )

            self.runner = TPUModelRunner(vllm_config,
                                         devices=self.mock_devices)

    def test_get_supported_tasks_runner(self):
        """Test get_supported_tasks for generate runner type."""
        supported_tasks = self.runner.get_supported_tasks()
        assert supported_tasks == ("generate", )

    def test_get_input_ids_embeds(self):
        """Tests _get_input_ids_embeds for both multimodal and text-only models."""
        # 1. ===== Setup =====
        dummy_input_ids = jnp.array([1, 2, 3])
        dummy_mm_embeds = jnp.ones((10, 128))
        dummy_is_mm_embed = jnp.array([False, True, True], dtype=jnp.bool_)
        dummy_final_embeds = jnp.ones((3, 128))

        # Mock the embedding function
        self.mock_get_input_embed_fn = MagicMock()
        self.runner.embed_input_ids_fn = self.mock_get_input_embed_fn
        self.mock_get_input_embed_fn.return_value = dummy_final_embeds
        self.runner.state = MagicMock()

        # 2. ===== Act & Assert (Multimodal) =====
        self.runner.is_multimodal_model = True

        input_ids_res, inputs_embeds_res = self.runner._get_input_ids_embeds(
            dummy_input_ids, dummy_mm_embeds, dummy_is_mm_embed)

        assert input_ids_res is None
        np.testing.assert_array_equal(np.asarray(inputs_embeds_res),
                                      np.asarray(dummy_final_embeds))
        self.mock_get_input_embed_fn.assert_called_once_with(
            self.runner.state,
            dummy_input_ids,
            dummy_mm_embeds,
            is_multimodal=dummy_is_mm_embed)

        # 3. ===== Act & Assert (Multimodal w/o mm embeds) =====
        self.mock_get_input_embed_fn.reset_mock()
        self.runner.is_multimodal_model = True

        # Without mm_embeds in the current scheduled tokens
        input_ids_res, inputs_embeds_res = self.runner._get_input_ids_embeds(
            dummy_input_ids, None, None)

        assert inputs_embeds_res is None
        np.testing.assert_array_equal(np.asarray(input_ids_res),
                                      np.asarray(dummy_input_ids))
        self.mock_get_input_embed_fn.assert_not_called()

        # 4. ===== Act & Assert (Text-only) =====
        self.mock_get_input_embed_fn.reset_mock()
        self.runner.is_multimodal_model = False

        input_ids_res, inputs_embeds_res = self.runner._get_input_ids_embeds(
            dummy_input_ids, dummy_mm_embeds, dummy_is_mm_embed)

        assert inputs_embeds_res is None
        np.testing.assert_array_equal(np.asarray(input_ids_res),
                                      np.asarray(dummy_input_ids))
        self.mock_get_input_embed_fn.assert_not_called()

    @patch('tpu_inference.runner.tpu_runner.TPUSupportedSamplingMetadata')
    def test_prepare_inputs_hybrid_kvcache(self, mock_sampling_metadata):
        # create hybrid kv cache config
        # 20 layers, 10 full attn + 10 sw attn
        self._create_mock_hybrid_kv_cache_config()

        # Mock scheduler output.
        scheduler_output = MagicMock()
        scheduler_output.total_num_scheduled_tokens = 10
        scheduler_output.num_scheduled_tokens = {'req1': 10}
        scheduler_output.scheduled_spec_decode_tokens = {}
        scheduler_output.grammar_bitmask = None

        # Mock input_batch
        self.runner.input_batch = MagicMock()
        self.runner.input_batch.num_reqs = 1
        self.runner.input_batch.req_ids = ['req1']
        self.runner.input_batch.req_id_to_index = {'req1': 0}
        self.runner.input_batch.num_computed_tokens_cpu = np.array([10])
        self.runner.input_batch.token_ids_cpu = np.random.randint(
            0, 1000, (8, 64), dtype=np.int32)

        # Mock block tables
        # there will be 2 block tables since there are 2 kv cache groups
        mock_block_table = MagicMock()
        mock_block_table.get_cpu_tensor.return_value = np.zeros(
            self.runner.block_tables_cpu[0].shape)
        self.runner.input_batch.block_table = [
            mock_block_table, mock_block_table
        ]
        self.runner.block_tables_cpu = [
            np.zeros(self.runner.block_tables_cpu[0].shape, dtype=np.int32),
            np.zeros(self.runner.block_tables_cpu[0].shape, dtype=np.int32)
        ]

        mock_sampling_instance = MagicMock()
        mock_sampling_metadata.from_input_batch.return_value = mock_sampling_instance

        output = self.runner._prepare_inputs_non_dp(scheduler_output)
        assert len(output) == 8
        input_ids, positions, attention_metadata, sampling_metadata, logits_indices, spec_decode_metadata, logits_indices_selector, padded_num_reqs = output
        # assert it will create attention metadata for each layer.
        assert isinstance(attention_metadata, dict)
        assert len(attention_metadata) == 20

    def _create_mock_hybrid_kv_cache_config(self):
        mock_kv_cache_config = MagicMock()
        mock_kv_cache_group1 = MagicMock()
        mock_kv_cache_group1.layer_names = [f'layer.{i}' for i in range(10)]
        mock_kv_cache_group2 = MagicMock()
        mock_kv_cache_group2.layer_names = [
            f'layer.{i}' for i in range(10, 20)
        ]
        mock_kv_cache_config.kv_cache_groups = [
            mock_kv_cache_group1, mock_kv_cache_group2
        ]
        self.runner.kv_cache_config = mock_kv_cache_config
        self.runner.use_hybrid_kvcache = True


class TestTPUJaxRunnerMultimodalModelLoadedForTextOnly:

    def setup_method(self):
        # Mock JAX dependencies
        self.mock_devices = [MagicMock(coords=i) for i in range(4)]
        self.mock_rng_key = MagicMock()
        device_array = np.array(jax.devices()[:1]).reshape(1, 1, 1, -1)
        self.mock_mesh = jax.make_mesh(device_array.shape,
                                       ('data', 'attn_dp', 'expert', 'model'))
        # Setup the runner with the model_config.is_multimodal_model set to True but get_model returning None for embed_multimodal_fn and embed_input_ids_fn.
        with patch('jax.devices', return_value=self.mock_devices), \
             patch('jax.make_mesh', return_value=self.mock_mesh), \
             patch('jax.random.key', return_value=self.mock_rng_key), \
             patch('tpu_inference.runner.tpu_runner.nnx.Rngs', return_value=self.mock_rng_key), \
             patch('tpu_inference.runner.tpu_runner.get_model', return_value=self._model_get_model()), \
             patch('tpu_inference.runner.tpu_runner.make_optimized_mesh', return_value=self.mock_mesh):

            model_config = ModelConfig(tokenizer_mode="auto",
                                       trust_remote_code=False,
                                       seed=0,
                                       dtype='bfloat16')
            # Set multimodal_config to not None, such that the is_multimodal_model property of model_config is True.
            model_config.multimodal_config = MagicMock()

            cache_config = CacheConfig(
                block_size=16,
                gpu_memory_utilization=0.9,
                cache_dtype="auto",
            )
            scheduler_config = SchedulerConfig(max_num_seqs=16,
                                               max_model_len=1024,
                                               is_encoder_decoder=False)
            parallel_config = ParallelConfig(
                pipeline_parallel_size=1,
                tensor_parallel_size=1,
            )
            vllm_config = VllmConfig(
                model_config=model_config,
                cache_config=cache_config,
                scheduler_config=scheduler_config,
                parallel_config=parallel_config,
                speculative_config=None,
                observability_config={},
                additional_config={},
            )

            self.runner = TPUModelRunner(vllm_config,
                                         devices=self.mock_devices)
            self.runner.load_model()

    def _model_get_model(self):
        mock_multimodal_fns = {
            "precompile_vision_encoder_fn": None,
            "embed_multimodal_fn": None,
            "embed_input_ids_fn": None,
            "get_mrope_input_positions_fn": None
        }
        return (
            MagicMock(),  # TPUModelRunner.model_fn
            MagicMock(),  # TPUModelRunner.compute_logits_fn
            MagicMock(),  # TPUModelRunner.pooler_fn
            MagicMock(),  # TPUModelRunner.combine_hidden_states_fn
            mock_multimodal_fns,  # TPUModelRunner.multimodal_fns
            MagicMock(),  # TPUModelRunner.state (model params)
            None,  # TPUModelRunner.lora_manager
            None,  # TPUModelRunner.model
        )

    def test_is_multimodal_model(self):
        # Precondition: make sure the model_config claims the model supports MM.
        assert self.runner.model_config.is_multimodal_model

        # Precondition: load the model and returns embed_multimodal_fn as None.
        assert self.runner.embed_multimodal_fn is None

        assert not self.runner.is_multimodal_model

        self.runner.embed_input_ids_fn = MagicMock()
        dummy_input_ids = jnp.array([1, 2, 3])
        dummy_mm_embeds = jnp.ones((10, 128))
        dummy_is_mm_embed = jnp.array([False, True, True], dtype=jnp.bool_)
        _ = self.runner._get_input_ids_embeds(dummy_input_ids, dummy_mm_embeds,
                                              dummy_is_mm_embed)
        self.runner.embed_input_ids_fn.assert_not_called()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import MagicMock, patch

from vllm.distributed.kv_transfer.kv_connector.v1.base import KVConnectorRole
from vllm.v1.request import RequestStatus

from tpu_inference.distributed import tpu_connector


class MockVllmConfig:

    def __init__(self):
        self.kv_transfer_config = MagicMock()
        self.kv_transfer_config.is_kv_producer = True
        self.cache_config = MagicMock()
        self.cache_config.block_size = 16
        self.parallel_config = MagicMock()


@patch("tpu_inference.distributed.tpu_connector.TPUConnectorWorker")
@patch("tpu_inference.distributed.tpu_connector.TPUConnectorScheduler")
class TestTPUConnector(unittest.TestCase):

    def setUp(self):
        self.vllm_config = MockVllmConfig()

    def test_init_scheduler_role(self, mock_scheduler_cls, mock_worker_cls):
        """
        Tests that TPUConnector initializes the scheduler connector for the
        SCHEDULER role.
        """
        connector = tpu_connector.TPUConnector(self.vllm_config,
                                               KVConnectorRole.SCHEDULER)
        mock_scheduler_cls.assert_called_once_with(self.vllm_config)
        mock_worker_cls.assert_not_called()
        self.assertIsNotNone(connector.connector_scheduler)
        self.assertIsNone(connector.connector_worker)

    def test_init_worker_role(self, mock_scheduler_cls, mock_worker_cls):
        """
        Tests that TPUConnector initializes the worker connector for the WORKER
        role.
        """
        connector = tpu_connector.TPUConnector(self.vllm_config,
                                               KVConnectorRole.WORKER)
        mock_worker_cls.assert_called_once_with(self.vllm_config)
        mock_scheduler_cls.assert_not_called()
        self.assertIsNone(connector.connector_scheduler)
        self.assertIsNotNone(connector.connector_worker)

    def test_scheduler_methods_are_called(self, mock_scheduler_cls,
                                          mock_worker_cls):
        """Tests that scheduler-side methods are correctly delegated."""
        mock_scheduler_instance = mock_scheduler_cls.return_value
        connector = tpu_connector.TPUConnector(self.vllm_config,
                                               KVConnectorRole.SCHEDULER)

        mock_request = MagicMock()
        mock_blocks = MagicMock()
        mock_scheduler_output = MagicMock()

        connector.get_num_new_matched_tokens(mock_request, 16)
        mock_scheduler_instance.get_num_new_matched_tokens.assert_called_once_with(
            mock_request, 16)

        connector.update_state_after_alloc(mock_request, mock_blocks, 16)
        mock_scheduler_instance.update_state_after_alloc.assert_called_once_with(
            mock_request, mock_blocks, 16)

        connector.build_connector_meta(mock_scheduler_output)
        mock_scheduler_instance.build_connector_meta.assert_called_once_with()

        connector.request_finished(mock_request, [1, 2])
        mock_scheduler_instance.request_finished.assert_called_once_with(
            mock_request, [1, 2])

    def test_worker_methods_are_called(self, mock_scheduler_cls,
                                       mock_worker_cls):
        """Tests that worker-side methods are correctly delegated."""
        mock_worker_instance = mock_worker_cls.return_value
        connector = tpu_connector.TPUConnector(self.vllm_config,
                                               KVConnectorRole.WORKER)
        connector._connector_metadata = tpu_connector.TPUConnectorMetadata(
        )  # need to set this for start_load_kv

        mock_runner = MagicMock()

        connector.register_runner(mock_runner)
        mock_worker_instance.register_runner.assert_called_once_with(
            mock_runner)

        connector.start_load_kv(None)
        mock_worker_instance.process_send_load.assert_called_once_with(
            connector._connector_metadata)

        connector.get_finished(set())
        mock_worker_instance.get_finished.assert_called_once_with()


class TestTPUConnectorScheduler(unittest.TestCase):

    def setUp(self):
        self.vllm_config = MockVllmConfig()
        self.vllm_config.cache_config.block_size = 16
        self.vllm_config.kv_transfer_config.is_kv_producer = False

        with patch(
                "tpu_inference.distributed.tpu_connector.dist_utils.get_kv_ips",
                return_value="1.1.1.1"
        ), patch(
                "tpu_inference.distributed.tpu_connector.dist_utils.get_kv_ports",
                return_value=12345):
            self.scheduler = tpu_connector.TPUConnectorScheduler(
                self.vllm_config)

    def test_get_num_new_matched_tokens_producer(self):
        """Tests that producer returns 0 tokens to load."""
        self.scheduler.is_producer = True
        mock_request = MagicMock()
        num_tokens, is_async = self.scheduler.get_num_new_matched_tokens(
            mock_request, 16)
        self.assertEqual(num_tokens, 0)
        self.assertFalse(is_async)

    def test_get_num_new_matched_tokens_consumer_needs_loading(self):
        """Tests consumer calculates correct number of tokens to load."""
        mock_request = MagicMock()
        mock_request.prompt_token_ids = [0] * 35  # 2 blocks worth, plus some
        num_computed_tokens = 16  # 1 block
        # rounded_down(35) = 32. 32 - 16 = 16.
        expected_tokens = 16
        num_tokens, is_async = self.scheduler.get_num_new_matched_tokens(
            mock_request, num_computed_tokens)
        self.assertEqual(num_tokens, expected_tokens)
        self.assertTrue(is_async)

    def test_get_num_new_matched_tokens_consumer_no_loading(self):
        """Tests consumer returns 0 if prompt is fully cached."""
        mock_request = MagicMock()
        mock_request.prompt_token_ids = [0] * 31  # less than 2 blocks
        num_computed_tokens = 32  # 2 blocks computed
        expected_tokens = 0
        num_tokens, is_async = self.scheduler.get_num_new_matched_tokens(
            mock_request, num_computed_tokens)
        self.assertEqual(num_tokens, expected_tokens)
        self.assertFalse(is_async)

    def test_update_state_after_alloc_producer(self):
        """Tests that update_state_after_alloc is a no-op for producers."""
        self.scheduler.is_producer = True
        self.scheduler.update_state_after_alloc(MagicMock(), MagicMock(), 16)
        self.assertEqual(len(self.scheduler.reqs_to_load), 0)

    def test_update_state_after_alloc_consumer_with_external_tokens(self):
        """
        Tests consumer state is updated when external tokens are needed.
        """
        mock_request = MagicMock()
        mock_request.request_id = "req1"
        mock_request.kv_transfer_params = {
            "uuid": 123,
            "remote_block_ids": [10, 11],
            "remote_host": "2.2.2.2",
            "remote_port": 54321
        }
        mock_blocks = MagicMock()
        mock_blocks.get_block_ids.return_value = [[1, 2]]
        num_external_tokens = 32

        self.scheduler.update_state_after_alloc(mock_request, mock_blocks,
                                                num_external_tokens)

        self.assertIn("req1", self.scheduler.reqs_to_load)
        load_meta = self.scheduler.reqs_to_load["req1"]
        self.assertEqual(load_meta.uuid, 123)
        self.assertEqual(load_meta.local_block_ids, [1, 2])
        self.assertEqual(load_meta.remote_block_ids, [10, 11])

    def test_update_state_after_alloc_consumer_no_external_tokens(self):
        """
        Tests consumer state is updated for notification when no external
        tokens are needed.
        """
        mock_request = MagicMock()
        mock_request.request_id = "req1"
        mock_request.kv_transfer_params = {
            "uuid": 123,
            "remote_block_ids": [10, 11],
            "remote_host": "2.2.2.2",
            "remote_port": 54321
        }
        mock_blocks = MagicMock()
        num_external_tokens = 0

        self.scheduler.update_state_after_alloc(mock_request, mock_blocks,
                                                num_external_tokens)

        self.assertIn("req1", self.scheduler.reqs_to_load)
        load_meta = self.scheduler.reqs_to_load["req1"]
        self.assertEqual(load_meta.uuid, 123)
        self.assertIsNone(load_meta.local_block_ids)
        self.assertIsNone(load_meta.remote_block_ids)

    def test_build_connector_meta(self):
        """Tests that metadata is built and state is cleared."""
        self.scheduler.is_producer = True
        self.scheduler.reqs_to_send = {"req1": "meta1"}
        meta = self.scheduler.build_connector_meta()
        self.assertEqual(meta.reqs_to_send, {"req1": "meta1"})
        self.assertEqual(len(self.scheduler.reqs_to_send),
                         0)  # check it was cleared

        self.scheduler.is_producer = False
        self.scheduler.reqs_to_load = {"req2": "meta2"}
        meta = self.scheduler.build_connector_meta()
        self.assertEqual(meta.reqs_to_load, {"req2": "meta2"})
        self.assertEqual(len(self.scheduler.reqs_to_load), 0)

    def test_request_finished_consumer(self):
        """Tests request_finished is a no-op for consumers."""
        self.scheduler.is_producer = False
        delay_free, params = self.scheduler.request_finished(MagicMock(), [])
        self.assertFalse(delay_free)
        self.assertIsNone(params)

    @patch("tpu_inference.distributed.tpu_connector.get_uuid",
           return_value=456)
    def test_request_finished_producer_finished_by_length(self, mock_get_uuid):
        """Tests producer logic when a request finishes normally."""
        self.scheduler.is_producer = True
        mock_request = MagicMock()
        mock_request.request_id = "req-finished"
        mock_request.status = RequestStatus.FINISHED_LENGTH_CAPPED
        mock_request.num_computed_tokens = 32  # 2 blocks
        block_ids = [1, 2]

        delay_free, params = self.scheduler.request_finished(
            mock_request, block_ids)

        self.assertTrue(delay_free)
        self.assertIn("req-finished", self.scheduler.reqs_to_send)
        send_meta = self.scheduler.reqs_to_send["req-finished"]
        self.assertEqual(send_meta.uuid, 456)
        self.assertEqual(send_meta.local_block_ids, [1, 2])

        self.assertIsNotNone(params)
        self.assertEqual(params["uuid"], 456)
        self.assertEqual(params["remote_block_ids"], [1, 2])
        self.assertEqual(params["remote_host"], "1.1.1.1")
        self.assertEqual(params["remote_port"], 12345)

    def test_request_finished_producer_not_finished(self):
        """Tests producer logic when a request is not yet finished."""
        self.scheduler.is_producer = True
        mock_request = MagicMock()
        mock_request.status = RequestStatus.RUNNING  # Not finished
        delay_free, params = self.scheduler.request_finished(
            mock_request, [1, 2])
        self.assertFalse(delay_free)
        self.assertIsNone(params)

    def test_request_finished_producer_prompt_too_short(self):
        """Tests producer logic when prompt is too short to transfer."""
        self.scheduler.is_producer = True
        mock_request = MagicMock()
        mock_request.request_id = "req-short"
        mock_request.status = RequestStatus.FINISHED_LENGTH_CAPPED
        mock_request.num_computed_tokens = 10  # less than a block
        block_ids = [1]

        delay_free, params = self.scheduler.request_finished(
            mock_request, block_ids)

        self.assertFalse(delay_free)
        self.assertEqual(params, {})
        self.assertNotIn("req-short", self.scheduler.reqs_to_send)


class TestTPUConnectorWorker(unittest.TestCase):

    def setUp(self):
        self.vllm_config = MockVllmConfig()
        patchers = {
            "jax":
            patch('tpu_inference.distributed.tpu_connector.jax'),
            "get_host_ip":
            patch(
                'tpu_inference.distributed.tpu_connector.dist_utils.get_host_ip',
                return_value='127.0.0.1'),
            "get_kv_transfer_port":
            patch(
                'tpu_inference.distributed.tpu_connector.dist_utils.get_kv_transfer_port',
                return_value=10000),
            "get_side_channel_port":
            patch(
                'tpu_inference.distributed.tpu_connector.dist_utils.get_side_channel_port',
                return_value=20000),
            "start_transfer_server":
            patch(
                'tpu_inference.distributed.tpu_connector.start_transfer_server'
            ),
            "zmq":
            patch('tpu_inference.distributed.tpu_connector.zmq'),
            "threading":
            patch('tpu_inference.distributed.tpu_connector.threading'),
            "ThreadPoolExecutor":
            patch(
                'tpu_inference.distributed.tpu_connector.ThreadPoolExecutor'),
            "device_array":
            patch('tpu_inference.distributed.tpu_connector.device_array'),
            "select_from_kv_caches":
            patch(
                'tpu_inference.distributed.tpu_connector.select_from_kv_caches'
            ),
            "insert_kv_chunks":
            patch('tpu_inference.distributed.tpu_connector.insert_kv_chunks'),
            "time":
            patch('tpu_inference.distributed.tpu_connector.time'),
            "make_zmq_path":
            patch('tpu_inference.distributed.tpu_connector.make_zmq_path'),
            "make_zmq_socket":
            patch('tpu_inference.distributed.tpu_connector.make_zmq_socket'),
        }
        self.all_mocks = {k: p.start() for k, p in patchers.items()}
        self.all_mocks["jax"].local_devices.return_value = [MagicMock()]
        for p in patchers.values():
            self.addCleanup(p.stop)

    def test_init_producer(self):
        """Tests worker initialization for the producer role."""
        self.vllm_config.kv_transfer_config.is_kv_producer = True
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)

        self.all_mocks["zmq"].Context.assert_called_once()
        self.all_mocks["threading"].Thread.assert_called_once()
        self.all_mocks["threading"].Event.assert_called()
        self.all_mocks["ThreadPoolExecutor"].assert_not_called()
        self.assertTrue(worker.is_producer)

    def test_init_consumer(self):
        """Tests worker initialization for the consumer role."""
        self.vllm_config.kv_transfer_config.is_kv_producer = False
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)

        self.all_mocks["zmq"].Context.assert_called_once()
        self.all_mocks["threading"].Thread.assert_not_called()
        self.all_mocks["ThreadPoolExecutor"].assert_called_once_with(
            max_workers=128)
        self.assertFalse(worker.is_producer)

    def test_register_runner(self):
        """Tests that runner registration correctly sets worker attributes."""
        self.vllm_config.kv_transfer_config.is_kv_producer = False
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)

        mock_runner = MagicMock()
        mock_kv_cache_layer = MagicMock()
        mock_kv_cache_layer.shape = [10, 20, 30, 40]
        mock_kv_cache_layer.dtype = 'float32'
        mock_sharding = MagicMock()
        mock_sharding.mesh = 'mesh'
        mock_sharding.spec = 'sharding_spec'
        mock_kv_cache_layer.sharding = mock_sharding
        mock_runner.kv_caches = [mock_kv_cache_layer] * 5
        mock_runner.mesh = 'mesh'

        worker.register_runner(mock_runner)

        self.all_mocks["start_transfer_server"].assert_called_once()
        self.assertEqual(worker.runner, mock_runner)
        self.assertEqual(worker.mesh, 'mesh')
        self.assertEqual(worker.num_layers, 5)
        self.assertEqual(worker.shape, [10, 20, 30, 40])
        self.assertEqual(worker.dtype, 'float32')
        self.assertEqual(worker.sharding, mock_sharding)

    def test_process_send_load_for_producer(self):
        """Tests process_send_load for the producer role."""
        self.vllm_config.kv_transfer_config.is_kv_producer = True
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)
        worker._prepare_kv_and_wait = MagicMock()

        meta = tpu_connector.TPUConnectorMetadata()
        send_meta = tpu_connector.SendMeta(uuid=1,
                                           local_block_ids=[1],
                                           expiration_time=123)
        meta.reqs_to_send = {"req1": send_meta}

        worker.process_send_load(meta)

        worker._prepare_kv_and_wait.assert_called_once_with("req1", send_meta)

    def test_process_send_load_for_consumer_loading(self):
        """Tests process_send_load for a consumer that needs to load KV."""
        self.vllm_config.kv_transfer_config.is_kv_producer = False
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)
        worker._maybe_build_kv_connection = MagicMock(return_value="conn")
        mock_indices = "mocked_indices"
        self.all_mocks["device_array"].return_value = mock_indices

        meta = tpu_connector.TPUConnectorMetadata()
        load_meta = tpu_connector.LoadMeta(uuid=1,
                                           local_block_ids=[1],
                                           remote_block_ids=[10],
                                           remote_host="host",
                                           remote_port=123)
        meta.reqs_to_load = {"req1": load_meta}

        worker.process_send_load(meta)

        worker._maybe_build_kv_connection.assert_called_once_with(load_meta)
        self.all_mocks[
            "ThreadPoolExecutor"].return_value.submit.assert_called_once_with(
                worker._pull_kv, "req1", "conn", load_meta, mock_indices)

    def test_process_send_load_for_consumer_notifying(self):
        """Tests process_send_load for a consumer that needs to notify."""
        self.vllm_config.kv_transfer_config.is_kv_producer = False
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)
        worker._maybe_build_notif_socket = MagicMock(return_value="socket")
        worker._notify_pull_done = MagicMock()
        uuid = 10
        meta = tpu_connector.TPUConnectorMetadata()
        load_meta = tpu_connector.LoadMeta(uuid=uuid,
                                           local_block_ids=None,
                                           remote_block_ids=None,
                                           remote_host="host",
                                           remote_port=123)
        meta.reqs_to_load = {"req1": load_meta}

        worker.runner = MagicMock()
        original_kv_caches = worker.runner.kv_caches
        worker.sharding = MagicMock()
        worker.sharding.spec = "mock_spec"
        worker.mesh = "mock_mesh"
        worker.reqs_ready_to_insert = {"req1": ("kv_data", "indices", [1])}
        self.all_mocks['insert_kv_chunks'].return_value = "new_kv_caches"

        worker.process_send_load(meta)

        self.all_mocks['insert_kv_chunks'].assert_called_once_with(
            original_kv_caches, "kv_data", [1], "mock_mesh", "mock_spec")
        self.assertEqual(worker.runner.kv_caches, "new_kv_caches")
        self.assertNotIn("req1", worker.reqs_ready_to_insert)
        worker._maybe_build_notif_socket.assert_called_once_with(load_meta)
        worker._notify_pull_done.assert_called_once_with(
            "socket", "req1", uuid)

    def test_get_finished_recving(self):
        """Tests get_finished for a request that has finished pulling."""
        self.vllm_config.kv_transfer_config.is_kv_producer = False
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)
        worker.runner = MagicMock()

        mock_future = MagicMock()
        mock_future.done.return_value = True
        mock_future.result.return_value = ('kv_data', 'indices', [1])
        worker.reqs_pulling = {'req1': mock_future}

        done_sending, done_recving = worker.get_finished()

        self.assertEqual(done_sending, set())
        self.assertEqual(done_recving, {'req1'})
        self.assertNotIn('req1', worker.reqs_pulling)
        self.assertIn('req1', worker.reqs_ready_to_insert)
        self.assertEqual(worker.reqs_ready_to_insert['req1'],
                         ('kv_data', 'indices', [1]))
        self.all_mocks['insert_kv_chunks'].assert_not_called()

    def test_get_finished_sending_expired(self):
        """Tests get_finished for a request that has expired."""
        self.vllm_config.kv_transfer_config.is_kv_producer = True
        worker = tpu_connector.TPUConnectorWorker(self.vllm_config)

        self.all_mocks['time'].perf_counter.return_value = 1000
        worker.reqs_wait_pull = {'req1': ['kv_data', 900]}

        done_sending, done_recving = worker.get_finished()

        self.assertEqual(done_sending, {'req1'})
        self.assertEqual(done_recving, set())
        self.assertNotIn('req1', worker.reqs_wait_pull)


class TestTPUConnectorUtils(unittest.TestCase):

    @patch("tpu_inference.distributed.tpu_connector.multi_layer_copy")
    def test_insert_kv_chunks_unmocked_contiguous(self, mock_multi_layer_copy):
        """
        Tests that insert_kv_chunks groups contiguous block numbers into a single
        chunk when calling multi_layer_copy.
        """
        from tpu_inference.distributed.tpu_connector import insert_kv_chunks

        # We need mock arrays that have .shape attribute
        y_slice = MagicMock()
        y_slice.shape = [4, 10]
        kv_slices = [y_slice]
        kv_caches = [MagicMock()]
        mesh = MagicMock()
        spec = MagicMock()

        block_numbers = [5, 6, 7, 8, 10, 11, 17, 18, 19]

        insert_kv_chunks(kv_caches, kv_slices, block_numbers, mesh, spec)

        mock_multi_layer_copy.assert_called_once()
        _, kwargs = mock_multi_layer_copy.call_args

        import numpy as np
        self.assertEqual(kwargs['dest_offsets'].shape[0], 9)
        np.testing.assert_array_equal(kwargs['dest_offsets'][:3], [5, 10, 17])
        self.assertEqual(kwargs['chunk_sizes'].shape[0], 9)
        np.testing.assert_array_equal(kwargs['chunk_sizes'][:3], [4, 2, 3])
        np.testing.assert_array_equal(kwargs['num_chunks'], [3])


if __name__ == "__main__":
    unittest.main()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import functools
from enum import Enum
from typing import Any

import jax
import jax.numpy as jnp
from jax._src import dtypes
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu


class RpaCase(Enum):
    """Represents the different cases for Ragged Paged Attention.

  - DECODE: Sequences are in decode-only mode (q_len = 1).
  - PREFILL: Sequences are in prefill-only mode (q_len > 1, static).
  - MIXED: Sequences can be a mix of prefill and decode (q_len > 1, dynamic).
  """

    DECODE = 0
    PREFILL = 1
    MIXED = 2

    @property
    def symbol(self):
        return {
            RpaCase.DECODE: "d",
            RpaCase.PREFILL: "p",
            RpaCase.MIXED: "m",
        }[self]

    def get_range(self, distribution):
        assert distribution.shape == (3, )
        if self == RpaCase.DECODE:
            return 0, distribution[0]
        elif self == RpaCase.PREFILL:
            return distribution[0], distribution[1]
        elif self == RpaCase.MIXED:
            return distribution[1], distribution[2]
        else:
            raise ValueError(f"Unsupported RPA case: {self}")


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True, eq=True)
class RPAConfig:
    num_seq: int
    bq_sz: int
    bkv_sz: int
    batch_size: int
    page_size: int
    bkv_p: int
    pages_per_seq: int
    max_steps_ub: int
    total_q_tokens: int
    total_num_pages: int | None = None
    head_dim: int = 128
    num_kv_heads: int = 8
    sm_scale: float = 1.0
    soft_cap: Any = None
    num_q_heads_per_kv_head: int = 1
    sliding_window: Any = None
    mask_value: float = -1e30
    q_dtype: Any = jnp.bfloat16
    kv_dtype: Any = jnp.bfloat16
    q_scale: Any = None
    k_scale: Any = None
    v_scale: Any = None
    vmem_limit_bytes: int | None = None
    case: RpaCase = RpaCase.MIXED
    n_buffer: int = 2
    out_dtype: Any = jnp.bfloat16
    fuse_accum: bool = False
    mask_v: bool = True

    @property
    def bkv_p_cache(self) -> int:
        if self.case == RpaCase.PREFILL:
            return 0
        return self.bkv_p

    @property
    def bkv_p_new(self) -> int:
        if self.case == RpaCase.DECODE:
            return 1
        return self.bkv_p

    @property
    def num_page_indices(self) -> int:
        return self.num_seq * self.pages_per_seq

    @property
    def page_size_log2(self) -> int:
        return (self.page_size - 1).bit_length()

    @property
    def page_size_mask(self) -> int:
        return self.page_size - 1

    @property
    def int_ty(self) -> Any:
        if (get_dtype_packing(self.q_dtype) != 1
                and pltpu.get_tpu_info().generation >= 6):
            return jnp.int16
        return jnp.int32


def get_dtype_bitwidth(dtype):
    return dtypes.itemsize_bits(dtype)


def get_dtype_packing(dtype):
    bits = get_dtype_bitwidth(dtype)
    return 32 // bits


def has_bank_conflicts(stride: int, distance=24, num_banks=32) -> bool:
    banks = set()
    for i in range(distance):
        bank = (i * stride) % num_banks
        if bank in banks:
            return True
        banks.add(bank)
    return False


@jax.tree_util.register_pytree_node_class
@dataclasses.dataclass(frozen=True)
class RPASchedule:
    """Container for metadata arrays with integrated shape/spec logic."""

    s_idx: Any
    q_idx: Any
    k_idx: Any
    is_last_k: Any  # [steps * batch]
    do_writeback: Any  # [steps * batch]
    dma_q: Any  # [steps * batch * 2]
    dma_kv_cache: Any  # [steps * batch * bkv_p_cache * 3]
    dma_kv_new: Any  # [steps * batch * bkv_p_new * 4]
    actual_steps: Any
    batch_size: int = dataclasses.field(default=0, metadata={"static": True})
    bkv_p_cache: int = dataclasses.field(default=0, metadata={"static": True})
    bkv_p_new: int = dataclasses.field(default=0, metadata={"static": True})

    def tree_flatten(self):
        return (
            self.s_idx,
            self.q_idx,
            self.k_idx,
            self.is_last_k,
            self.do_writeback,
            self.dma_q,
            self.dma_kv_cache,
            self.dma_kv_new,
            self.actual_steps,
        ), (self.batch_size, self.bkv_p_cache, self.bkv_p_new)

    @classmethod
    def tree_unflatten(cls, aux, children):
        return cls(
            *children,
            batch_size=aux[0],
            bkv_p_cache=aux[1],
            bkv_p_new=aux[2],
        )

    num_metadata_arrays = 9

    def get_dma_kv_cache(self, step: int, batch_idx: int,
                         page_idx: int) -> tuple[int, int, int]:
        # Stride is 3: src_hbm, dst_vmem, size
        cache_step_stride = self.batch_size * self.bkv_p_cache * 3
        cache_batch_stride = self.bkv_p_cache * 3
        idx = (step * cache_step_stride + batch_idx * cache_batch_stride +
               page_idx * 3)
        # 0: src_hbm, 1: dst_vmem, 2: size
        src_off = self.dma_kv_cache[idx + 0]
        dst_off = self.dma_kv_cache[idx + 1]
        sz = self.dma_kv_cache[idx + 2]
        return src_off, dst_off, sz

    def get_dma_kv_new(self, step: int, batch_idx: int,
                       page_idx: int) -> tuple[int, int, int, int]:
        # Stride is 4: dst_hbm, src_hbm, dst_vmem, size
        new_step_stride = self.batch_size * self.bkv_p_new * 4
        base_idx = (step * new_step_stride + batch_idx * self.bkv_p_new * 4 +
                    page_idx * 4)
        # 0: dst_hbm, 1: src_hbm, 2: dst_vmem, 3: size
        dst_hbm = self.dma_kv_new[base_idx + 0]
        src_hbm = self.dma_kv_new[base_idx + 1]
        dst_vmem = self.dma_kv_new[base_idx + 2]
        sz = self.dma_kv_new[base_idx + 3]
        return dst_hbm, src_hbm, dst_vmem, sz

    def get_dma_q(self, step: int, batch_idx: int) -> tuple[int, int]:
        # Stride is 2: src_hbm, size
        q_step_stride = self.batch_size * 2
        base_idx = step * q_step_stride + batch_idx * 2
        # 0: src_hbm, 1: size
        src_hbm = self.dma_q[base_idx + 0]
        sz = self.dma_q[base_idx + 1]
        return src_hbm, sz

    @classmethod
    def _map_shapes(cls, config: RPAConfig, wrapper):
        bs = config.batch_size
        steps = config.max_steps_ub
        return cls(
            s_idx=wrapper((steps * bs, ), is_smem=False),
            q_idx=wrapper((steps * bs, ), is_smem=False),
            k_idx=wrapper((steps * bs, ), is_smem=False),
            is_last_k=wrapper((steps * bs, ), is_smem=False),
            do_writeback=wrapper((steps * bs, ), is_smem=False),
            dma_q=wrapper((steps * bs * 2, ), is_smem=False),
            # we do this hacky workaround because xla throws an error if we have
            # zero-sized tensors.
            dma_kv_cache=wrapper(
                (max(1, steps * bs * config.bkv_p_cache * 3), ),
                is_smem=False),
            dma_kv_new=wrapper((max(1, steps * bs * config.bkv_p_new * 4), ),
                               is_smem=False),
            # this needs to be in SMEM when we prefetch it so we can read from it immediately
            actual_steps=wrapper((1, ), is_smem=True),
            batch_size=bs,
            bkv_p_cache=config.bkv_p_cache,
            bkv_p_new=config.bkv_p_new,
        )

    @classmethod
    def out_shape(cls, config: RPAConfig):
        """Returns a Pytree of ShapeDtypeStruct for pallas_call."""
        return cls._map_shapes(
            config,
            lambda shape, is_smem=False: jax.ShapeDtypeStruct(
                shape, jnp.int32),
        )

    @classmethod
    def hbm_specs(cls, config: RPAConfig):
        """Returns a Pytree of BlockSpecs in HBM matching the output structure."""

        def wrapper(shape, is_smem=False):
            return pl.BlockSpec(memory_space=pltpu.HBM, block_shape=shape)

        return cls._map_shapes(config, wrapper)

    @classmethod
    def kernel_in_specs(cls, config: RPAConfig):
        """Returns a Pytree of BlockSpecs for passing schedule to kernel."""

        def wrapper(shape, is_smem=False):
            memory_space = pltpu.SMEM if is_smem else pltpu.HBM
            return pl.BlockSpec(memory_space=memory_space, block_shape=shape)

        return cls._map_shapes(config, wrapper)

    @classmethod
    def out_specs(cls, config: RPAConfig):
        """Returns a Pytree of BlockSpecs matching the output structure."""

        def wrapper(shape, is_smem=False):
            return pl.BlockSpec(memory_space=pltpu.SMEM, block_shape=shape)

        return cls._map_shapes(config, wrapper)

    @classmethod
    def smem_specs(cls, config: RPAConfig):
        """Returns a Pytree of pltpu.SMEM matching the output structure."""
        return cls._map_shapes(
            config, lambda shape, is_smem=False: pltpu.SMEM(shape, jnp.int32))


def rpa_metadata_schedule_kernel(
    ## HBM inputs
    cu_q_lens_ref,
    kv_lens_ref,
    distribution_ref,
    # output
    schedule_hbm: RPASchedule,
    # scratch
    schedule: RPASchedule,
    lane_lengths_ref,
    dma_sem,
    *,
    config: RPAConfig,
):
    """Generates the HBM-to-VMEM DMA schedule

  This kernel:
  1. Iterates through each (potentially ragged) sequence
  2. Breaks Queries (Q) and Key-Values (KV) into blocks (bq_sz, bkv_sz).
  3. Assigns tasks to 'lanes' (TPU batch items) based on current lane occupancy
     to ensure balanced execution across the batch dimension.
  4. Encodes DMA offsets:
     - dma_q: HBM start index and size for Query blocks.
     - dma_kv_cache: Paged indices for existing KV tokens.
     - dma_kv_new: offsets for new tokens being added to the cache.
     - do_writeback: boolean flag indicating if a block should be flushed to
     HBM (ie does this block contain new tokens to add to KV cache).

  Check bref_override.py to check usage of the schedule.
  """
    for b in range(config.batch_size):
        lane_lengths_ref[b] = 0

    def seq_loop(s_idx, _):
        q_start = cu_q_lens_ref[s_idx]
        q_end = cu_q_lens_ref[s_idx + 1]
        k_len = kv_lens_ref[s_idx]
        q_len = q_end - q_start

        n_q = (q_len + config.bq_sz - 1) // config.bq_sz
        n_k = (k_len + config.bkv_sz - 1) // config.bkv_sz

        def q_loop(q_idx, _):
            target_lane = 0
            min_len = lane_lengths_ref[0]
            for b in range(1, config.batch_size):
                is_better = lane_lengths_ref[b] < min_len
                target_lane = jnp.where(is_better, b, target_lane)
                min_len = jnp.where(is_better, lane_lengths_ref[b], min_len)

            curr_ptr = lane_lengths_ref[target_lane]
            q_src = q_start + q_idx * config.bq_sz
            q_sz_task = jnp.clip(q_end - q_src, 0, config.bq_sz)

            start_k_idx = 0
            if config.sliding_window is not None:
                sw_start_idx = (k_len - q_len + q_idx * config.bq_sz -
                                config.sliding_window + 1)
                start_k_idx = jnp.maximum(0, sw_start_idx) // config.bkv_sz

            end_k_idx_causal = (k_len - q_len + q_idx * config.bq_sz +
                                q_sz_task - 1) // config.bkv_sz + 1
            end_k_idx = jnp.minimum(n_k, end_k_idx_causal)

            def k_loop(k_idx, curr_ptr):

                idx = curr_ptr * config.batch_size + target_lane
                schedule.s_idx[idx] = s_idx
                schedule.q_idx[idx] = q_idx
                schedule.k_idx[idx] = k_idx
                schedule.is_last_k[idx] = jnp.asarray(k_idx == end_k_idx - 1,
                                                      dtype=jnp.int32)

                q_idx_base = curr_ptr * (config.batch_size *
                                         2) + target_lane * 2
                schedule.dma_q[q_idx_base + 0] = q_src
                schedule.dma_q[q_idx_base + 1] = q_sz_task

                kv_len_start = k_idx * config.bkv_sz
                kv_p_start = k_idx * config.bkv_p
                kv_left = k_len - kv_len_start
                kv_left_frm_cache = jnp.maximum(kv_left - q_len, 0)
                p_offset = s_idx * config.pages_per_seq + kv_p_start

                kv_cache_base = curr_ptr * (
                    config.batch_size * config.bkv_p_cache *
                    3) + target_lane * (config.bkv_p_cache * 3)
                for i in range(config.bkv_p_cache):
                    sz = jnp.clip(
                        kv_left_frm_cache - (i << config.page_size_log2),
                        0,
                        config.page_size,
                    )
                    p_idx = jnp.minimum(p_offset + i,
                                        config.num_page_indices - 1)

                    base_i = kv_cache_base + i * 3
                    schedule.dma_kv_cache[base_i + 0] = p_idx
                    schedule.dma_kv_cache[base_i +
                                          1] = i << config.page_size_log2
                    schedule.dma_kv_cache[base_i + 2] = sz

                kv_left_frm_new = kv_left - kv_left_frm_cache
                bkv_sz_cache = jnp.minimum(kv_left_frm_cache, config.bkv_sz)
                new_sz = jnp.minimum(config.bkv_sz - bkv_sz_cache,
                                     kv_left_frm_new)

                # Writeback logic: each new k block is written back by the first q block that attends to it.
                q_wb = jnp.maximum(0, (k_idx * config.bkv_sz -
                                       (k_len - q_len)) // config.bq_sz)
                schedule.do_writeback[idx] = jnp.asarray(
                    (new_sz > 0) & (q_idx == q_wb), dtype=jnp.int32)

                kv_new_base = curr_ptr * (
                    config.batch_size * config.bkv_p_new *
                    4) + target_lane * (config.bkv_p_new * 4)

                if config.case == RpaCase.DECODE:
                    # During DECODE, we only have 1 new token, so we only need 1 page slot.
                    start_in_slot = bkv_sz_cache
                    sz = new_sz

                    p_idx = (kv_len_start +
                             start_in_slot) >> config.page_size_log2
                    p_idx = jnp.minimum(p_idx, config.pages_per_seq - 1)
                    p_off = (kv_len_start +
                             start_in_slot) & config.page_size_mask
                    global_p_idx = s_idx * config.pages_per_seq + p_idx

                    schedule.dma_kv_new[kv_new_base + 0] = (
                        global_p_idx << config.page_size_log2) | p_off
                    schedule.dma_kv_new[kv_new_base +
                                        1] = q_end - kv_left_frm_new
                    schedule.dma_kv_new[kv_new_base + 2] = start_in_slot
                    schedule.dma_kv_new[kv_new_base + 3] = sz
                else:
                    for i in range(config.bkv_p_new):
                        slot_start = i << config.page_size_log2
                        slot_end = (i + 1) << config.page_size_log2

                        start_in_slot = jnp.maximum(slot_start, bkv_sz_cache)
                        end_in_slot = jnp.minimum(slot_end,
                                                  bkv_sz_cache + new_sz)
                        sz = jnp.maximum(0, end_in_slot - start_in_slot)

                        p_idx = (kv_len_start +
                                 start_in_slot) >> config.page_size_log2
                        p_idx = jnp.minimum(p_idx, config.pages_per_seq - 1)
                        p_off = (kv_len_start +
                                 start_in_slot) & config.page_size_mask
                        global_p_idx = s_idx * config.pages_per_seq + p_idx

                        base_i = kv_new_base + i * 4
                        schedule.dma_kv_new[base_i + 0] = (
                            global_p_idx << config.page_size_log2) | p_off
                        schedule.dma_kv_new[base_i +
                                            1] = q_end - kv_left_frm_new
                        schedule.dma_kv_new[base_i + 2] = start_in_slot
                        schedule.dma_kv_new[base_i + 3] = sz
                return curr_ptr + 1

            lane_lengths_ref[target_lane] = jax.lax.fori_loop(
                start_k_idx, end_k_idx, k_loop, curr_ptr)
            return None

        jax.lax.fori_loop(0, n_q, q_loop, None)

    start_seq_idx, end_seq_idx = config.case.get_range(distribution_ref)
    jax.lax.fori_loop(start_seq_idx, end_seq_idx, seq_loop, None)

    max_steps = lane_lengths_ref[0]
    for b in range(1, config.batch_size):
        max_steps = jnp.where(lane_lengths_ref[b] > max_steps,
                              lane_lengths_ref[b], max_steps)
    schedule.actual_steps[0] = max_steps

    safe_max_steps = jnp.minimum(max_steps + config.n_buffer + 1,
                                 config.max_steps_ub)

    def mask_lane(b, _):
        start_step = lane_lengths_ref[b]

        def mask_step(step, _):
            idx = step * config.batch_size + b

            # Mark as invalid and zero out control flags
            schedule.s_idx[idx] = -1
            schedule.q_idx[idx] = 0
            schedule.k_idx[idx] = 0
            schedule.is_last_k[idx] = 0
            schedule.do_writeback[idx] = 0

            # TODO: theoretically we just need to zero out the size, check later
            q_base = step * config.batch_size * 2 + b * 2
            schedule.dma_q[q_base + 0] = 0
            schedule.dma_q[q_base + 1] = 0

            cache_base = (step * config.batch_size * config.bkv_p_cache * 3 +
                          b * config.bkv_p_cache * 3)
            for i in range(config.bkv_p_cache):
                base_i = cache_base + i * 3
                schedule.dma_kv_cache[base_i + 0] = 0
                schedule.dma_kv_cache[base_i + 1] = 0
                schedule.dma_kv_cache[base_i + 2] = 0

            new_base = (step * config.batch_size * config.bkv_p_new * 4 +
                        b * config.bkv_p_new * 4)
            for i in range(config.bkv_p_new):
                base_i = new_base + i * 4
                schedule.dma_kv_new[base_i + 0] = 0
                schedule.dma_kv_new[base_i + 1] = 0
                schedule.dma_kv_new[base_i + 2] = 0
                schedule.dma_kv_new[base_i + 3] = 0

        jax.lax.fori_loop(start_step, safe_max_steps, mask_step, None)

    jax.lax.fori_loop(0, config.batch_size, mask_lane, None)

    flat_hbm = jax.tree_util.tree_leaves(schedule_hbm)
    flat_smem = jax.tree_util.tree_leaves(schedule)
    dma_list = []
    sem = dma_sem.at[0]

    for h, s in zip(flat_hbm, flat_smem):
        if h.shape[0] != 1:
            multiplier = h.shape[0] // config.max_steps_ub
            fetch_size = multiplier * safe_max_steps
            # Ensure fetch_size is at most h.shape[0] to prevent out-of-bounds DMA
            fetch_size = pl.cdiv(fetch_size, 1024) * 1024

            copy = pltpu.make_async_copy(
                s.at[pl.ds(0, fetch_size)],
                h.at[pl.ds(0, fetch_size)],
                sem,
            )
            copy.start()
            dma_list.append(copy)
        else:
            copy = pltpu.make_async_copy(
                s,
                h,
                sem,
            )
            copy.start()
            dma_list.append(copy)

    for copy in dma_list:
        copy.wait()


def generate_rpa_metadata(
    cu_q_lens,
    kv_lens,
    distribution,
    config: RPAConfig,
    *,
    interpret=False,
):
    scratch_shapes = [
        RPASchedule.smem_specs(config),
        pltpu.SMEM((config.batch_size, ), jnp.int32),
        pltpu.SemaphoreType.DMA((1, )),
    ]

    return pl.pallas_call(
        functools.partial(rpa_metadata_schedule_kernel, config=config),
        out_shape=RPASchedule.out_shape(config),
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=3,
            in_specs=[],
            out_specs=RPASchedule.hbm_specs(config),
            scratch_shapes=scratch_shapes,
        ),
        interpret=interpret,
        name="rpa_metadata_schedule",
    )(cu_q_lens, kv_lens, distribution)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# /home/pooyam/tpu_inference/tests/models/jax/layers/test_sampling.py
import jax.numpy as jnp
import numpy as np
from vllm.v1.outputs import LogprobsTensors

from tpu_inference.layers.jax.sample.sampling import (compute_logprobs,
                                                      gather_logprobs)


class TestSampling:

    def test_compute_logprobs(self):
        logits = jnp.array([[1.0, 2.0, 3.0], [3.0, 2.0, 1.0]],
                           dtype=jnp.float32)
        logprobs = compute_logprobs(logits)

        # Expected values computed with scipy.special.log_softmax
        expected_logprobs = np.array(
            [
                [-2.40760596, -1.40760596, -0.40760596],
                [-0.40760596, -1.40760596, -2.40760596],
            ],
            dtype=np.float32,
        )
        assert np.allclose(logprobs, expected_logprobs, atol=1e-6)

    def test_gather_logprobs(self):
        logprobs = jnp.array(
            [
                [-2.40760596, -1.40760596, -0.40760596, -3.40760596],
                [-0.40760596, -1.40760596, -2.40760596, -3.40760596],
            ],
            dtype=jnp.float32,
        )
        token_ids = jnp.array([2, 0], dtype=jnp.int32)
        num_logprobs = 2

        result: LogprobsTensors = gather_logprobs(logprobs, token_ids,
                                                  num_logprobs)

        # check indices
        expected_indices = np.array(
            [
                [2, 2, 1],  # token id 2, top-k are 2, 1
                [0, 0, 1],  # token id 0, top-k are 0, 1
            ],
            dtype=np.int32,
        )
        assert np.array_equal(result.logprob_token_ids, expected_indices)

        # check logprobs
        expected_logprobs_values = np.array(
            [
                [-0.40760596, -0.40760596, -1.40760596],
                [-0.40760596, -0.40760596, -1.40760596],
            ],
            dtype=np.float32,
        )
        assert np.allclose(result.logprobs,
                           expected_logprobs_values,
                           atol=1e-6)

        # check ranks
        expected_ranks = np.array([1, 1], dtype=np.int32)
        assert np.array_equal(result.selected_token_ranks, expected_ranks)

    def test_gather_logprobs_with_ties(self):
        logprobs = jnp.array(
            [
                [-1.0, -1.0, -2.0, -2.0],
            ],
            dtype=jnp.float32,
        )
        token_ids = jnp.array([1], dtype=jnp.int32)
        num_logprobs = 3

        result: LogprobsTensors = gather_logprobs(logprobs, token_ids,
                                                  num_logprobs)

        # check logprobs
        expected_logprobs_values = np.array(
            [
                [-1.0, -1.0, -1.0, -2.0],
            ],
            dtype=np.float32,
        )
        assert np.allclose(result.logprobs,
                           expected_logprobs_values,
                           atol=1e-6)

        # check ranks
        # rank of token 1 is 2 because there are 2 values >= -1.0
        expected_ranks = np.array([2], dtype=np.int32)
        assert np.array_equal(result.selected_token_ranks, expected_ranks)

        # check indices
        # The order of tied elements is not guaranteed.
        # token id is 1. top-k indices are a permutation of {0, 1, 2} or {0, 1, 3}.
        assert result.logprob_token_ids[0, 0] == 1
        top_k_indices = sorted(result.logprob_token_ids[0, 1:].tolist())
        assert top_k_indices == [0, 1, 2] or top_k_indices == [0, 1, 3]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, fields

import jax
import torch
from jax._src import mesh as meshlib
from jax.sharding import Mesh, NamedSharding, PartitionSpec
from torch.nn import ParameterList
from torch.nn.parameter import Parameter
from torchax.tensor import Tensor

from tpu_inference.layers.common.utils import (
    general_device_put, reorder_concatenated_tensor_for_sharding)
from tpu_inference.logger import init_logger

P = PartitionSpec

logger = init_logger(__name__)


@jax.tree_util.register_dataclass
@dataclass
class LinearWeights:
    weight: jax.Array | Tensor | list[jax.Array | Tensor]
    weight_scale: jax.Array | Tensor | list[jax.Array | Tensor] | None
    zero_point: jax.Array | Tensor | list[jax.Array | Tensor] | None
    bias: jax.Array | Tensor | list[jax.Array | Tensor] | None


MODEL_MATMUL_FUSION_TRUTH_TABLE = {
    ("Qwen/Qwen2.5-7B-Instruct", 1024, 1, "QKVParallelLinear"):
    True,
    ("Qwen/Qwen2.5-7B-Instruct", 1024, 1, "MergedColumnParallelLinear"):
    False,
    ("Qwen/Qwen2.5-7B-Instruct", 2048, 1, "QKVParallelLinear"):
    False,
    ("Qwen/Qwen2.5-7B-Instruct", 2048, 1, "MergedColumnParallelLinear"):
    False,
    ("meta-llama/Llama-3.1-8B-Instruct", 1024, 1, "QKVParallelLinear"):
    False,
    ("meta-llama/Llama-3.1-8B-Instruct", 1024, 1, "MergedColumnParallelLinear"):
    False,
    ("meta-llama/Llama-3.1-8B-Instruct", 2048, 1, "QKVParallelLinear"):
    False,
    ("meta-llama/Llama-3.1-8B-Instruct", 2048, 1, "MergedColumnParallelLinear"):
    False,
    ("RedHatAI/Meta-Llama-3.1-8B-Instruct-quantized.w8a8", 1024, 1, "QKVParallelLinear"):
    False,
    ("RedHatAI/Meta-Llama-3.1-8B-Instruct-quantized.w8a8", 1024, 1, "MergedColumnParallelLinear"):
    False,
    ("RedHatAI/Meta-Llama-3.1-8B-Instruct-quantized.w8a8", 2048, 1, "QKVParallelLinear"):
    False,
    ("RedHatAI/Meta-Llama-3.1-8B-Instruct-quantized.w8a8", 2048, 1, "MergedColumnParallelLinear"):
    False,
}


def to_parameter_list(tensor: list[torch.Tensor]):
    tensor = [Parameter(t, requires_grad=False) for t in tensor]
    return ParameterList(tensor)


def get_model_matmul_fusion_assignment(model_name: str, batch_size: int,
                                       tp_size: int, layer_name: str):
    key = (model_name, batch_size, tp_size, layer_name)
    return MODEL_MATMUL_FUSION_TRUTH_TABLE.get(key, True)


def process_linear_weights(
    weights: LinearWeights,
    fused: bool = False,
    output_sizes: list[int] | None = None,
    reorder_size: int | None = None,
    transposed: bool = True,
    per_tensor: bool = False,
) -> LinearWeights:
    weight = weights.weight
    weight_scale = weights.weight_scale
    zero_point = weights.zero_point
    bias = weights.bias

    dim = 0 if transposed else -1
    if output_sizes is None:
        output_sizes = [weight.shape[dim]]

    if fused:
        assert reorder_size is not None
        weight = reorder_concatenated_tensor_for_sharding(
            weight, output_sizes, reorder_size, dim)

        if weight_scale is not None and not per_tensor:
            weight_scale = reorder_concatenated_tensor_for_sharding(
                weight_scale, output_sizes, reorder_size, dim)
        if zero_point is not None:
            zero_point = reorder_concatenated_tensor_for_sharding(
                zero_point, output_sizes, reorder_size, dim)
        if bias is not None:
            bias = reorder_concatenated_tensor_for_sharding(
                bias, output_sizes, reorder_size, dim)
    else:

        def slice_tensor(tensor):
            tensors = []
            start = 0
            for size in output_sizes:
                end = start + size
                tensor_split = jax.lax.slice_in_dim(tensor,
                                                    start,
                                                    end,
                                                    axis=dim)
                tensors.append(tensor_split)
                start = end
            return tensors

        weight = slice_tensor(weight)
        if weight_scale is not None and not per_tensor:
            weight_scale = slice_tensor(weight_scale)
        if zero_point is not None:
            zero_point = slice_tensor(zero_point)
        if bias is not None:
            bias = slice_tensor(bias)

    return LinearWeights(
        weight=weight,
        weight_scale=weight_scale,
        zero_point=zero_point,
        bias=bias,
    )


def shard_linear_weights(
    weights: LinearWeights,
    mesh: Mesh | None,
    weight_p_spec: PartitionSpec,
    bias_p_spec: PartitionSpec,
    transposed: bool = True,
    per_tensor: bool = False,
) -> LinearWeights:
    # jax==0.8.1 introduces jax.sharding.get_mesh(), but current
    # v6e test environment uses 0.8.0, so we use jax._src.mesh instead.
    mesh = mesh or meshlib.get_concrete_mesh()
    if not transposed:
        # By defualt, we use transposed weights. If it is not transposed,
        # we need to transpose the sharding as well.
        weight_p_spec = PartitionSpec(*weight_p_spec[::-1])
        bias_p_spec = PartitionSpec(weight_p_spec[0])

    weight_sharding = NamedSharding(mesh, weight_p_spec)
    bias_sharding = NamedSharding(mesh, bias_p_spec)
    if isinstance(weights.weight_scale, (jax.Array, Tensor)) and len(
            weights.weight_scale.shape) == 3:
        num_blocks = weights.weight_scale.shape[0]
        if len(weight_p_spec) != 2:
            raise ValueError(
                F"The weight sharding shape length should be 2, but given {len(weight_p_spec)}."
            )
        # Cannot be sharded on the first dimension in case the number of blocks is 1.
        in_axis = weight_p_spec[1] if num_blocks > 1 else None
        out_axis = weight_p_spec[0]
        weight_scale_p_spec = P(in_axis, None, out_axis)
        weight_scale_sharding = NamedSharding(mesh, weight_scale_p_spec)
    else:
        weight_scale_sharding = NamedSharding(
            mesh, P()) if per_tensor else bias_sharding

    weight_shardings = LinearWeights(
        weight=weight_sharding,
        weight_scale=weight_scale_sharding,
        zero_point=bias_sharding,
        bias=bias_sharding,
    )

    for field in fields(LinearWeights):
        key = field.name
        if (weight := getattr(weights, key, None)) is not None:
            sharding = getattr(weight_shardings, key)
            weight = general_device_put(weight, sharding)
            setattr(weights, key, weight)
    return weights



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import MagicMock, patch

from tpu_inference.layers.common.sharding import (LazyShardingAxisName,
                                                  ShardingAxisName2D,
                                                  ShardingAxisNameBase,
                                                  ShardingConfigManager)


class TestShardingConfigManager(unittest.TestCase):

    @patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN", True)
    def test_sharding_config_manager_from_vllm_config(self):
        vllm_config = MagicMock()
        vllm_config.parallel_config.tensor_parallel_size = 8
        vllm_config.parallel_config.data_parallel_size = 2
        vllm_config.model_config.use_mla = True
        vllm_config.model_config.get_total_num_kv_heads.return_value = 1
        vllm_config.speculative_config = None
        vllm_config.lora_config = None

        # Test default sharding strategy
        vllm_config.additional_config = {"sharding": {}}

        manager = ShardingConfigManager.from_vllm_config(vllm_config)
        self.assertEqual(manager.tp_size, 8)
        self.assertEqual(manager.model_dp_size, 2)
        self.assertEqual(manager.attn_dp_size, 1)
        self.assertEqual(manager.attn_dp_expert_size, 1)
        self.assertEqual(manager.total_dp_size, 2)

    @patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN", True)
    def test_sharding_config_manager_with_dp_attention(self):
        vllm_config = MagicMock()
        vllm_config.parallel_config.tensor_parallel_size = 8
        vllm_config.parallel_config.data_parallel_size = 1
        vllm_config.model_config.use_mla = True
        vllm_config.model_config.get_total_num_kv_heads.return_value = 1
        vllm_config.speculative_config = None
        vllm_config.lora_config = None
        vllm_config.cache_config.cache_dtype = "bfloat16"

        # Test with enable_dp_attention=True and expert_parallelism
        vllm_config.additional_config = {
            "sharding": {
                "sharding_strategy": {
                    "enable_dp_attention": True,
                    "expert_parallelism": 4
                }
            }
        }

        manager = ShardingConfigManager.from_vllm_config(vllm_config)
        # num_kv_heads = 1 (MLA), packing = 2 (BF16)
        # num_kv_heads_per_device_in_kv_cache = max(1, (1 * 2) / 2) = 1
        # attn_dp = max(int(8 // 1), 1) = 8
        # tensor_parallelism = 8 // 8 = 1
        # attn_dp_expert = expert_parallelism = 4
        # expert_parallelism = 1

        self.assertEqual(manager.tp_size, 1)
        self.assertEqual(manager.attn_dp_size, 8)
        self.assertEqual(manager.attn_dp_expert_size, 4)
        self.assertEqual(manager.expert_size, 1)
        self.assertEqual(manager.total_dp_size, 32)  # 1 * 8 * 4

    def test_sharding_config_manager_explicit_tp(self):
        vllm_config = MagicMock()
        vllm_config.parallel_config.tensor_parallel_size = 1
        vllm_config.parallel_config.data_parallel_size = 1
        vllm_config.model_config.use_mla = True
        vllm_config.speculative_config = None

        vllm_config.additional_config = {
            "sharding": {
                "sharding_strategy": {
                    "tensor_parallelism": 8
                }
            }
        }

        manager = ShardingConfigManager.from_vllm_config(vllm_config)
        self.assertEqual(manager.tp_size, 8)

    def test_sharding_config_manager_tp_none(self):
        vllm_config = MagicMock()
        vllm_config.parallel_config.tensor_parallel_size = 8
        vllm_config.parallel_config.data_parallel_size = 1
        vllm_config.model_config.use_mla = True
        vllm_config.speculative_config = None

        # Test when sharding_strategy tensor_parallelism is None
        vllm_config.additional_config = {
            "sharding": {
                "sharding_strategy": {
                    "tensor_parallelism": None
                }
            }
        }

        manager = ShardingConfigManager.from_vllm_config(vllm_config)
        # Should fallback to pc_tensor_parallelism (8)
        self.assertEqual(manager.tp_size, 8)

    def test_sharding_config_manager_tp_equal_1(self):
        vllm_config = MagicMock()
        vllm_config.parallel_config.tensor_parallel_size = 8
        vllm_config.parallel_config.data_parallel_size = 1
        vllm_config.model_config.use_mla = True
        vllm_config.speculative_config = None

        # Test when sharding_strategy tensor_parallelism is 1
        vllm_config.additional_config = {
            "sharding": {
                "sharding_strategy": {
                    "tensor_parallelism": 1
                }
            }
        }

        manager = ShardingConfigManager.from_vllm_config(vllm_config)
        # Should use ss_tensor_parallelism (1)
        self.assertEqual(manager.tp_size, 1)

    def test_sharding_config_manager_tp_different(self):
        vllm_config = MagicMock()
        vllm_config.parallel_config.tensor_parallel_size = 8
        vllm_config.parallel_config.data_parallel_size = 1
        vllm_config.model_config.use_mla = True
        vllm_config.speculative_config = None

        # Test when sharding_strategy tensor_parallelism is different from pc_tensor_parallelism
        vllm_config.additional_config = {
            "sharding": {
                "sharding_strategy": {
                    "tensor_parallelism": 4
                }
            }
        }

        manager = ShardingConfigManager.from_vllm_config(vllm_config)
        # Should use ss_tensor_parallelism (4)
        self.assertEqual(manager.tp_size, 4)


class TestLazyShardingAxisName(unittest.TestCase):

    def test_initial_state_is_uninitialized(self):
        lazy = LazyShardingAxisName()
        self.assertIsNone(lazy._cls)

    @patch("tpu_inference.layers.common.sharding.envs.USE_2D_TP", True)
    @patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN", False)
    def test_use_2d_tp_selects_base(self):
        lazy = LazyShardingAxisName()
        _ = lazy.SEQUENCE
        self.assertIs(lazy._cls, ShardingAxisNameBase)

    @patch("tpu_inference.layers.common.sharding.envs.USE_2D_TP", False)
    @patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN", False)
    def test_both_false_selects_2d(self):
        lazy = LazyShardingAxisName()
        _ = lazy.SEQUENCE
        self.assertIs(lazy._cls, ShardingAxisName2D)

    def test_cls_cached_after_first_access(self):
        lazy = LazyShardingAxisName()
        with patch("tpu_inference.layers.common.sharding.envs.USE_2D_TP",
                   False), \
             patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN",
                   False):
            _ = lazy.SEQUENCE
            first_cls = lazy._cls

        # Even with different env vars, _cls should not change
        with patch("tpu_inference.layers.common.sharding.envs.USE_2D_TP",
                   True), \
             patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN",
                   True):
            _ = lazy.SEQUENCE
            self.assertIs(lazy._cls, first_cls)

    def test_initialized_after_env_change_uses_new_value(self):
        # LazyShardingAxisName is created while USE_2D_TP=False,
        # then USE_2D_TP changes to True before the first attribute access.
        # It should pick up the value at initialization time (first access),
        # not at construction time.
        with patch("tpu_inference.layers.common.sharding.envs.USE_2D_TP",
                   False), \
             patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN",
                   False):
            lazy = LazyShardingAxisName()
            self.assertIsNone(lazy._cls)  # not yet initialized

        with patch("tpu_inference.layers.common.sharding.envs.USE_2D_TP",
                   True), \
             patch("tpu_inference.layers.common.sharding.envs.NEW_MODEL_DESIGN",
                   False):
            _ = lazy.SEQUENCE  # initialized here, after env changed
            self.assertIs(lazy._cls, ShardingAxisNameBase)


if __name__ == "__main__":
    unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from jax.sharding import Mesh
from transformers import Qwen3Config
from vllm.config import VllmConfig

from tpu_inference import envs, utils
from tpu_inference.distributed.jax_parallel_state import get_pp_group
from tpu_inference.layers.common.attention_interface import attention
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.common.quantization import quantize_kv
from tpu_inference.layers.jax import JaxModule
from tpu_inference.layers.jax.embed import JaxEmbed
from tpu_inference.layers.jax.linear import JaxEinsum
from tpu_inference.layers.jax.norm import JaxRmsNorm
from tpu_inference.layers.jax.pp_utils import PPMissingLayer, make_layers
from tpu_inference.layers.jax.rope_interface import apply_rope
from tpu_inference.layers.vllm.quantization.configs import VllmQuantConfig
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.jax_intermediate_tensor import \
    JaxIntermediateTensors
from tpu_inference.models.jax.qwen2 import Qwen2DecoderLayer
from tpu_inference.models.jax.qwen2 import Qwen2MLP as Qwen3MLP
from tpu_inference.models.jax.qwen2 import Qwen2Model
from tpu_inference.models.jax.utils.weight_utils import LoadableWithIterator

logger = init_logger(__name__)

init_fn = nnx.initializers.uniform()


class Qwen3Attention(JaxModule):

    def __init__(self,
                 config: Qwen3Config,
                 dtype: jnp.dtype,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 kv_cache_dtype: str,
                 quant_config: VllmQuantConfig,
                 prefix: str = ""):
        self.hidden_size = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.num_kv_heads = config.num_key_value_heads
        self.rope_theta = config.rope_theta
        self.rope_scaling = getattr(config, "rope_scaling", None)
        self.rms_norm_eps = config.rms_norm_eps

        self.head_dim_original = getattr(config, "head_dim",
                                         self.hidden_size // self.num_heads)
        self.head_dim = utils.get_padded_head_dim(self.head_dim_original)

        sharding_size = mesh.shape["model"]
        self.num_heads = utils.get_padded_num_heads(self.num_heads,
                                                    sharding_size)
        self.num_kv_heads = utils.get_padded_num_heads(self.num_kv_heads,
                                                       sharding_size)

        self.mesh = mesh

        # NOTE: LAYOUT_Q_PROJ_AS_NDH is by default False
        if envs.LAYOUT_Q_PROJ_AS_NDH:
            rhs_str = "NDH"
            q_proj_sharding = ("model", None, None)
            kernel_shape = (self.num_heads, self.hidden_size, self.head_dim)
        else:
            rhs_str = "DNH"
            q_proj_sharding = (None, "model", None)
            kernel_shape = (self.hidden_size, self.num_heads, self.head_dim)

        logger.info_once(
            f"Running with attention Q-Projection laid out as {rhs_str}")

        self.q_proj = JaxEinsum(
            f"TD,{rhs_str}->TNH",
            kernel_shape,
            dtype=dtype,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, q_proj_sharding),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".q_proj",
        )
        self.q_norm = JaxRmsNorm(
            self.head_dim,
            epsilon=self.rms_norm_eps,
            dtype=dtype,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".q_norm",
        )
        self.k_proj = JaxEinsum(
            "TD,DKH->TKH",
            (self.hidden_size, self.num_kv_heads, self.head_dim),
            dtype=dtype,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, (None, "model", None)),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".k_proj",
        )
        self.k_norm = JaxRmsNorm(
            self.head_dim,
            epsilon=self.rms_norm_eps,
            dtype=dtype,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".k_norm",
        )
        self.v_proj = JaxEinsum(
            "TD,DKH->TKH",
            (self.hidden_size, self.num_kv_heads, self.head_dim),
            dtype=dtype,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, (None, "model", None)),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".v_proj",
        )
        self.o_proj = JaxEinsum(
            "TNH,NHD->TD",
            (self.num_heads, self.head_dim, self.hidden_size),
            dtype=dtype,
            param_dtype=dtype,
            kernel_init=nnx.with_partitioning(init_fn, ("model", None, None)),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".o_proj",
        )

        self._q_scale = 1.0
        self._k_scale = 1.0
        self._v_scale = 1.0
        self.kv_cache_quantized_dtype = None
        if kv_cache_dtype != "auto":
            self.kv_cache_quantized_dtype = utils.get_jax_dtype_from_str_dtype(
                kv_cache_dtype)

    def __call__(
        self,
        kv_cache: Optional[jax.Array],
        x: jax.Array,
        attention_metadata: AttentionMetadata,
    ) -> Tuple[jax.Array, jax.Array]:
        md = attention_metadata
        # q: (T, N, H)
        q = self.q_proj(x)
        q = self.q_norm(q)
        q = apply_rope(q, md.input_positions, self.head_dim_original,
                       self.rope_theta, self.rope_scaling)

        # k: (T, K, H)
        k = self.k_proj(x)
        k = self.k_norm(k)
        k = apply_rope(k, md.input_positions, self.head_dim_original,
                       self.rope_theta, self.rope_scaling)

        # v: (T, K, H)
        v = self.v_proj(x)
        # o: (T, N, H)
        q_scale = k_scale = v_scale = None
        if self.kv_cache_quantized_dtype:
            # TODO(kyuyeunk/jacobplatin): Enable w8a8 when VREG spill issue is resolved.
            # q_scale = self._q_scale
            k_scale = self._k_scale
            v_scale = self._v_scale
            k, v = quantize_kv(self.kv_cache_quantized_dtype, k, v, k_scale,
                               v_scale)

        new_kv_cache, outputs = attention(
            kv_cache,
            q,
            k,
            v,
            attention_metadata,
            self.mesh,
            self.head_dim_original,
            q_scale=q_scale,
            k_scale=k_scale,
            v_scale=v_scale,
        )
        # (T, D)
        o = self.o_proj(outputs)
        return new_kv_cache, o


class Qwen3DecoderLayer(Qwen2DecoderLayer):

    def __init__(self,
                 config: Qwen3Config,
                 dtype: jnp.dtype,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 kv_cache_dtype: str,
                 quant_config: VllmQuantConfig,
                 prefix: str = ""):
        rms_norm_eps = config.rms_norm_eps
        hidden_size = config.hidden_size

        self.input_layernorm = JaxRmsNorm(
            hidden_size,
            epsilon=rms_norm_eps,
            dtype=dtype,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".input_layernorm",
        )
        self.self_attn = Qwen3Attention(config=config,
                                        dtype=dtype,
                                        rng=rng,
                                        mesh=mesh,
                                        kv_cache_dtype=kv_cache_dtype,
                                        quant_config=quant_config,
                                        prefix=prefix + ".self_attn")
        self.post_attention_layernorm = JaxRmsNorm(
            hidden_size,
            epsilon=rms_norm_eps,
            dtype=dtype,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".post_attention_layernorm",
        )
        self.mlp = Qwen3MLP(
            config=config,
            dtype=dtype,
            rng=rng,
            quant_config=quant_config,
            prefix=prefix + ".mlp",
        )


class Qwen3Model(Qwen2Model):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 prefix: str = "model") -> None:
        model_config = vllm_config.model_config
        hf_config = model_config.hf_config
        vocab_size = model_config.get_vocab_size()
        dtype = model_config.dtype
        rms_norm_eps = hf_config.rms_norm_eps
        hidden_size = hf_config.hidden_size

        self.is_first_rank = get_pp_group().is_first_rank
        self.is_last_rank = get_pp_group().is_last_rank

        if self.is_first_rank or (hf_config.tie_word_embeddings
                                  and self.is_last_rank):
            self.embed_tokens = JaxEmbed(
                num_embeddings=vocab_size,
                features=hidden_size,
                dtype=dtype,
                param_dtype=dtype,
                embedding_init=nnx.with_partitioning(init_fn, ("model", None)),
                rngs=rng,
                quant_config=vllm_config.quant_config,
                prefix=prefix + ".embed_tokens",
            )
        else:
            self.embed_tokens = PPMissingLayer()

        self.start_layer, self.end_layer, self.layers = make_layers(
            hf_config.num_hidden_layers,
            lambda layer_index: Qwen3DecoderLayer(
                config=hf_config,
                dtype=dtype,
                rng=rng,
                mesh=mesh,
                # TODO (jacobplatin): we should refactor this to pass a dtype (or config) directly
                kv_cache_dtype=vllm_config.cache_config.cache_dtype,
                quant_config=vllm_config.quant_config,
                prefix=f"{prefix}.layers.{layer_index}",
            ))
        if self.is_last_rank:
            self.norm = JaxRmsNorm(
                hidden_size,
                epsilon=rms_norm_eps,
                dtype=dtype,
                param_dtype=dtype,
                scale_init=nnx.with_partitioning(init_fn, (None, )),
                rngs=rng,
                quant_config=vllm_config.quant_config,
                prefix=prefix + ".norm",
            )
        else:
            self.norm = PPMissingLayer()


class Qwen3ForCausalLM(JaxModule, LoadableWithIterator):

    def __init__(self, vllm_config: VllmConfig, rng_key: jax.Array,
                 mesh: Mesh) -> None:
        if getattr(vllm_config.model_config, "quantization", None) == "fp8":
            # `get_tpu_quantization_config` returns None for "fp8" because
            # the work in #1623 is not fully merged. So this block overrides
            # the logic to return Fp8Config when model_config indicates fp8.
            # TODO(#1623): Remove this block when `get_tpu_quantization_config`
            # is updated.
            from tpu_inference.layers.jax.quantization.fp8 import Fp8Config
            hg_quant_config = getattr(vllm_config.model_config.hf_config,
                                      "quantization_config", {})
            vllm_config.quant_config = Fp8Config(hg_quant_config)

        self.vllm_config = vllm_config
        rng = nnx.Rngs(rng_key)
        self.mesh = mesh

        self.model = Qwen3Model(
            vllm_config=vllm_config,
            rng=rng,
            mesh=mesh,
            prefix="model",
        )
        model_config = vllm_config.model_config
        if not model_config.hf_config.tie_word_embeddings:
            if self.model.is_last_rank:
                vocab_size = model_config.get_vocab_size()
                hidden_size = model_config.hf_config.hidden_size
                self.lm_head = JaxEinsum(
                    einsum_str="TD,DV->TV",
                    kernel_shape=(hidden_size, vocab_size),
                    dtype=model_config.dtype,
                    param_dtype=model_config.dtype,
                    rngs=rng,
                    quant_config=vllm_config.quant_config,
                    prefix="lm_head",
                )
            else:
                self.lm_head = PPMissingLayer()

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
        _input_positions=None,
        _layer_name_to_kv_cache=None,
        _lora_metadata=None,
        intermediate_tensors: JaxIntermediateTensors | None = None,
        is_first_rank: bool = True,
        is_last_rank: bool = True,
        *args,
    ) -> Tuple[List[jax.Array], jax.Array | JaxIntermediateTensors,
               List[jax.Array]]:
        if not is_first_rank:
            assert intermediate_tensors is not None
            inputs_embeds = intermediate_tensors["hidden_states"]
        kv_caches, x = self.model(
            kv_caches,
            input_ids,
            attention_metadata,
            inputs_embeds,
        )
        if not is_last_rank:
            x = JaxIntermediateTensors(tensors={"hidden_states": x}, )
        return kv_caches, x, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        if hasattr(self, 'lm_head'):
            return self.lm_head(hidden_states)

        assert isinstance(self.model.embed_tokens, JaxEmbed)
        return self.model.embed_tokens.decode(hidden_states)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch
from torchax.interop import jax_view, torch_view
from vllm.model_executor.layers.fused_moe import FusedMoE, FusedMoEMethodBase
from vllm.model_executor.layers.fused_moe.config import FusedMoEConfig

from tpu_inference import envs
from tpu_inference.layers.common.moe import MoEBackend, moe_apply
from tpu_inference.layers.common.process_weights.moe_weights import \
    FusedMoEWeights
from tpu_inference.logger import init_logger

logger = init_logger(__name__)


def select_moe_backend_from_fused_moe_config(
        moe: FusedMoEConfig) -> MoEBackend:
    """
    Select the MoE backend based on the FusedMoEConfig.

    NOTE (jacobplatin): we don't currently support DENSE_MAT or MEGABLX_GMM
    backends on the vLLM path for now.

    Args:
        moe: The FusedMoEConfig.

    Returns:
        The selected MoE backend.
    """

    if envs.USE_MOE_EP_KERNEL:
        if moe.use_ep:
            logger.info_once("[MoE]: Using fused MoE EP kernel")
            return MoEBackend.FUSED_MOE
        logger.warning_once(
            "USE_MOE_EP_KERNEL=1 but expert parallelism is not "
            "enabled. Falling back to gmm implementation.")

    if moe.use_ep:
        logger.info_once("[MoE]: Using GMM EP kernel")
        return MoEBackend.GMM_EP

    # Use default implementation.
    logger.info_once("[MoE]: Using GMM TP kernel")
    return MoEBackend.GMM_TP


def vllm_moe_apply(layer: FusedMoE, weights: FusedMoEWeights,
                   quant_method_instance: FusedMoEMethodBase, x: torch.Tensor,
                   router_logits: torch.Tensor) -> torch.Tensor:
    """
    Shared function for applying a FusedMoE layer for the TorchAX/vLLM backend.

    Args:
        layer: The FusedMoE layer.
        weights: The FusedMoE weights.
        quant_method_instance: The quantization method instance.
        x: The input tensor.
        router_logits: The router logits.

    Returns:
        The output tensor from the MoE fowrard pass.
    """
    assert isinstance(layer, FusedMoE)
    assert isinstance(quant_method_instance, FusedMoEMethodBase)
    assert isinstance(weights, FusedMoEWeights)

    return torch_view(
        moe_apply(
            layer=layer,
            x=jax_view(x),
            gating_output=jax_view(router_logits),
            weights=weights,
            moe_backend=quant_method_instance.moe_backend,
            mesh=quant_method_instance.mesh,
            extra_backend_kwargs=quant_method_instance.extra_backend_kwargs,
        ))



# SPDX-License-Identifier: Apache-2.0
# Test for LoRA weight loading API

import os
import tempfile
from dataclasses import dataclass
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from flax import nnx
from jax._src import test_util as jtu
from jax.sharding import Mesh
from safetensors.numpy import save_file

from tpu_inference.models.jax.utils.weight_utils import (
    MetadataMap, load_hf_weights, transfer_state_with_mappings)

# ----- nnx.Module Wrappers -----


class SourceLayer(nnx.Module):

    def __init__(self, rngs):
        self.kernel = nnx.Param(jax.random.normal(rngs(), (4, 4)))
        self.bias = nnx.Param(jax.random.normal(rngs(), (4, )))


class SourceModel(nnx.Module):

    def __init__(self, rngs):
        self.src_lm_head = nnx.Param(jax.random.normal(rngs(), (2, 4)))
        self.layers = nnx.Dict({'0': SourceLayer(rngs)})


class TargetLinear(nnx.Module):

    def __init__(self, rngs):
        self.kernel = nnx.Param(jnp.zeros((4, 4)))
        self.bias = nnx.Param(jnp.zeros((4, )))


class TargetBlock(nnx.Module):

    def __init__(self, rngs):
        self.mlp = nnx.Dict({"up_proj": TargetLinear(rngs)})


class TargetModel(nnx.Module):

    def __init__(self, rngs):
        self.tgt_lm_head = nnx.Param(jnp.zeros((2, 4)))
        self.model = nnx.Dict({"layers": nnx.Dict({'0': TargetBlock(rngs)})})


# ----- Test -----
class WeightTransfer(jtu.JaxTestCase):

    def test_transfer_state(self):
        rng = nnx.Rngs(0)
        src_model = SourceModel(rng)
        tgt_model = TargetModel(rng)

        # Get split states
        _, src_state = nnx.split(src_model)
        _, tgt_state = nnx.split(tgt_model)

        # Overwrite known values
        src_state["layers"]['0']["kernel"].value = jnp.ones((4, 4)) * 42.0
        src_state["layers"]['0']["bias"].value = jnp.ones((4, )) * 7.0
        src_state["src_lm_head"].value = jnp.ones((2, 4)) * 6.0
        # Mapping for both kernel and bias
        mappings = {
            "layers.*.kernel": ("model.layers.*.mlp.up_proj.kernel", (None, )),
            "layers.*.bias": ("model.layers.*.mlp.up_proj.bias", (None, )),
            "src_lm_head": ("tgt_lm_head", (None, None)),
        }

        # Transfer
        new_tgt_state = transfer_state_with_mappings(src_state, tgt_state,
                                                     mappings)

        # Assert correctness
        assert jnp.allclose(
            new_tgt_state["model"]["layers"]['0']["mlp"]["up_proj"]
            ["kernel"].value, 42.0)
        assert jnp.allclose(
            new_tgt_state["model"]["layers"]['0']["mlp"]["up_proj"]
            ["bias"].value, 7.0)
        assert jnp.allclose(new_tgt_state["tgt_lm_head"].value, 6.0)


# ----- Mocks for dtype test -----


class DtypeTestModel(nnx.Module):

    def __init__(self, dtype: jnp.dtype, rngs: nnx.Rngs):
        self.weight_to_cast = nnx.Param(jnp.zeros((2, 2), dtype=dtype))
        self.weight_to_keep = nnx.Param(jnp.zeros((2, 2), dtype=dtype))


@dataclass
class MockModelConfig:
    model: str
    dtype: jnp.dtype
    hf_config: Any = None

    def get_vocab_size(self):
        return 1

    def get_hidden_size(self):
        return 1

    def get_head_size(self):
        return 1

    is_multimodal_model: bool = False


@dataclass
class MockLoadConfig:
    download_dir: str


@dataclass
class MockVllmConfig:
    model_config: MockModelConfig
    load_config: MockLoadConfig
    speculative_config: Any = None


class WeightLoadingDtypeTest(jtu.JaxTestCase):

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tempdir.cleanup)

        # Create dummy safetensors file
        tensors = {
            "weight_to_cast.weight": np.ones((2, 2), dtype=np.float32),
            "weight_to_keep.weight": np.ones((2, 2), dtype=np.float32),
        }
        self.safetensors_path = os.path.join(self.tempdir.name,
                                             "model.safetensors")
        save_file(tensors, self.safetensors_path)

    def test_keep_original_dtype(self):
        rng = nnx.Rngs(0)
        model_dtype = jnp.bfloat16
        model = DtypeTestModel(dtype=model_dtype, rngs=rng)

        mock_model_config = MockModelConfig(model=self.tempdir.name,
                                            dtype=model_dtype)
        mock_load_config = MockLoadConfig(download_dir=self.tempdir.name)
        vllm_config = MockVllmConfig(model_config=mock_model_config,
                                     load_config=mock_load_config)

        mesh = Mesh(jax.devices(), ("model", ))

        name_map = {
            "weight_to_cast": "weight_to_cast",
            "weight_to_keep": "weight_to_keep",
        }
        metadata_map = MetadataMap(name_map=name_map)

        keep_original_dtype_keys_regex = [r"weight_to_keep.*"]

        load_hf_weights(
            vllm_config=vllm_config,
            model=model,
            metadata_map=metadata_map,
            mesh=mesh,
            keep_original_dtype_keys_regex=keep_original_dtype_keys_regex,
        )

        self.assertEqual(model.weight_to_cast.value.dtype, model_dtype)
        self.assertEqual(model.weight_to_keep.value.dtype, jnp.float32)
